import textwrap
from typing import Optional

import pytest

from debputy.highlevel_manifest import HighLevelManifest
from debputy.highlevel_manifest_parser import YAMLManifestParser
from debputy.interpreter import extract_shebang_interpreter
from debputy.plugin.api import virtual_path_def
from debputy.plugin.api.test_api import build_virtual_file_system
from debputy.transformation_rules import NormalizeShebangLineTransformation


@pytest.mark.parametrize(
    "raw_shebang,original_command,command_full_basename,command_stem,correct_command,corrected_shebang_line",
    [
        (
            b"#!       /usr/bin/false\r\n",
            "/usr/bin/false",
            "false",
            "false",
            None,
            None,
        ),
        (
            b"#!/usr/bin/python3 -b",
            "/usr/bin/python3",
            "python3",
            "python",
            "/usr/bin/python3",
            None,
        ),
        (
            b"#!/usr/bin/env python3 -b",
            "/usr/bin/env python3",
            "python3",
            "python",
            "/usr/bin/python3",
            "#! /usr/bin/python3 -b",
        ),
        (
            b"#! /bin/env python3.12-dbg -b",
            "/bin/env python3.12-dbg",
            "python3.12-dbg",
            "python",
            "/usr/bin/python3.12-dbg",
            "#! /usr/bin/python3.12-dbg -b",
        ),
        (
            b"#! /usr/bin/bash",
            "/usr/bin/bash",
            "bash",
            "bash",
            "/bin/bash",
            "#! /bin/bash",
        ),
        (
            b"#! /usr/bin/env sh",
            "/usr/bin/env sh",
            "sh",
            "sh",
            "/bin/sh",
            "#! /bin/sh",
        ),
        (
            b"#! /usr/local/bin/perl",
            "/usr/local/bin/perl",
            "perl",
            "perl",
            "/usr/bin/perl",
            "#! /usr/bin/perl",
        ),
    ],
)
def test_interpreter_detection(
    raw_shebang: bytes,
    original_command: str,
    command_full_basename: str,
    command_stem: str,
    correct_command: Optional[str],
    corrected_shebang_line: Optional[str],
) -> None:
    interpreter = extract_shebang_interpreter(raw_shebang)
    # The `and ...` part is just to get the raw line in the error message
    assert interpreter is not None

    assert interpreter.original_command == original_command
    assert interpreter.command_full_basename == command_full_basename
    assert interpreter.command_stem == command_stem
    assert interpreter.correct_command == correct_command
    assert interpreter.corrected_shebang_line == corrected_shebang_line
    assert interpreter.fixup_needed == (corrected_shebang_line is not None)


@pytest.mark.parametrize(
    "raw_data",
    [
        b"#!#!#!       /usr/bin/false",
        b"#!perl",  # Used in files as an editor hint
        b"\x7fELF/usr/bin/perl",
        b"\x00\01\x02\x03/usr/bin/perl",
        b"PK\x03\x03/usr/bin/perl",
    ],
)
def test_interpreter_negative(raw_data: bytes) -> None:
    assert extract_shebang_interpreter(raw_data) is None


@pytest.fixture
def empty_manifest(
    amd64_dpkg_architecture_variables,
    dpkg_arch_query,
    source_package,
    package_single_foo_arch_all_cxt_amd64,
    amd64_substitution,
    no_profiles_or_build_options,
    debputy_plugin_feature_set,
) -> HighLevelManifest:
    # We need an empty directory to avoid triggering packager provided files.
    debian_dir = build_virtual_file_system([])
    return YAMLManifestParser(
        "debian/test-debputy.manifest",
        source_package,
        package_single_foo_arch_all_cxt_amd64,
        amd64_substitution,
        amd64_dpkg_architecture_variables,
        dpkg_arch_query,
        no_profiles_or_build_options,
        debputy_plugin_feature_set,
        "full",
        debian_dir=debian_dir,
    ).build_manifest()


def test_interpreter_rewrite(empty_manifest: HighLevelManifest) -> None:
    condition_context = empty_manifest.condition_context("foo")
    fs_root = build_virtual_file_system(
        [
            virtual_path_def("usr/bin/foo", content="random data"),
            virtual_path_def(
                "usr/bin/foo.sh",
                materialized_content="#!/usr/bin/sh\nset -e\n",
            ),
        ]
    )
    interpreter_normalization = NormalizeShebangLineTransformation()
    interpreter_normalization.transform_file_system(fs_root, condition_context)
    foo = fs_root.lookup("usr/bin/foo")
    foo_sh = fs_root.lookup("usr/bin/foo.sh")

    assert foo is not None and foo.is_file
    with foo.open() as fd:
        assert fd.read() == "random data"

    assert foo_sh is not None and foo_sh.is_file
    with foo_sh.open() as fd:
        expected = textwrap.dedent(
            """\
        #! /bin/sh
        set -e
        """
        )
        assert fd.read() == expected



import os
import random
from typing import cast, TYPE_CHECKING, Sequence, Optional, Mapping

import pytest

from debputy.packager_provided_files import detect_all_packager_provided_files
from debputy.plugin.api import DebputyPluginInitializer
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.plugin.api.impl import plugin_metadata_for_debputys_own_plugin
from debputy.plugin.api.impl_types import (
    PackagerProvidedFileClassSpec,
    DebputyPluginMetadata,
)
from debputy.plugin.api.test_api import (
    InitializedPluginUnderTest,
    build_virtual_file_system,
)
from debputy.plugin.api.test_api.test_impl import (
    initialize_plugin_under_test_preloaded,
)
from lint_tests.lint_tutil import requires_levenshtein
from tutil import faked_binary_package, binary_package_table

if TYPE_CHECKING:
    from debputy.plugin.api.test_api.test_impl import InitializedPluginUnderTestImpl

    # Irrelevant, but makes the import not "unused" for things that does not parse `cast("...", ...)` expressions
    assert InitializedPluginUnderTestImpl is not None


def ppf_test_plugin(api: DebputyPluginInitializer) -> None:
    api.packager_provided_file(
        "arch-specific-dash",
        "/some/test-directory/{name}.conf",
        allow_architecture_segment=True,
    )
    api.packager_provided_file(
        "arch.specific.dot",
        "/some/test.directory/{name}",
        allow_architecture_segment=True,
    )

    api.packager_provided_file(
        "arch.specific.with.priority",
        "/some/test.priority.directory/{priority:02}-{name}",
        allow_architecture_segment=True,
        default_priority=60,
    )

    api.packager_provided_file(
        "packageless-fallback",
        "/some/test-plfb/{name}",
        packageless_is_fallback_for_all_packages=True,
    )
    api.packager_provided_file(
        "packageless.fallback",
        "/some/test.plfb/{name}",
        packageless_is_fallback_for_all_packages=True,
    )


@pytest.mark.parametrize(
    "package_name,basename,install_target,is_main_binary",
    [
        ("foo", "foo.arch-specific-dash", "./some/test-directory/foo.conf", True),
        # main package match
        ("foo", "arch-specific-dash", "./some/test-directory/foo.conf", True),
        # arch match
        ("foo", "foo.arch-specific-dash.amd64", "./some/test-directory/foo.conf", True),
        # Matches with periods in both package name and in the file type
        ("foo.bar", "foo.bar.arch.specific.dot", "./some/test.directory/foo.bar", True),
        ("foo.bar", "arch.specific.dot", "./some/test.directory/foo.bar", True),
        (
            "foo.bar",
            "foo.bar.arch.specific.dot.amd64",
            "./some/test.directory/foo.bar",
            True,
        ),
        # Priority
        (
            "foo.bar",
            "foo.bar.arch.specific.with.priority",
            "./some/test.priority.directory/60-foo.bar",
            True,
        ),
        (
            "foo.bar",
            "arch.specific.with.priority",
            "./some/test.priority.directory/60-foo.bar",
            True,
        ),
        (
            "foo.bar",
            "foo.bar.arch.specific.with.priority.amd64",
            "./some/test.priority.directory/60-foo.bar",
            True,
        ),
        # Name
        (
            "foo.bar",
            "foo.bar.special.name.arch.specific.with.priority",
            "./some/test.priority.directory/60-special.name",
            True,
        ),
        (
            "foo.bar",
            "foo.bar.special.name.arch.specific.with.priority.amd64",
            "./some/test.priority.directory/60-special.name",
            True,
        ),
        (
            "foo.bar",
            "packageless-fallback",
            "./some/test-plfb/foo.bar",
            False,
        ),
        (
            "foo.bar",
            "packageless.fallback",
            "./some/test.plfb/foo.bar",
            False,
        ),
    ],
)
def test_packager_provided_files(
    package_name: str, basename: str, install_target: str, is_main_binary: bool
) -> None:
    # Inject our custom files
    plugin = initialize_plugin_under_test_preloaded(
        1,
        ppf_test_plugin,
        plugin_name="pff-test-plugin",
    )
    debputy_plugin_feature_set = _fetch_debputy_plugin_feature_set(plugin)

    debian_dir = build_virtual_file_system([basename])
    binary_under_test = faked_binary_package(
        package_name, is_main_package=is_main_binary
    )
    main_package = (
        binary_under_test if is_main_binary else faked_binary_package("main-pkg")
    )
    binaries = [main_package]
    if not is_main_binary:
        binaries.append(binary_under_test)
    binary_packages = binary_package_table(*binaries)

    ppf = detect_all_packager_provided_files(
        debputy_plugin_feature_set.packager_provided_files,
        debian_dir,
        binary_packages,
    )
    assert package_name in ppf
    all_matched = ppf[package_name].auto_installable
    assert len(all_matched) == 1
    matched = all_matched[0]
    assert basename == matched.path.name
    actual_install_target = "/".join(matched.compute_dest())
    assert actual_install_target == install_target


@pytest.mark.parametrize(
    "package_name,expected_basename,non_matched",
    [
        ("foo", "foo.arch-specific-dash", ["arch-specific-dash"]),
        (
            "foo",
            "foo.arch-specific-dash.amd64",
            [
                "foo.arch-specific-dash",
                "arch-specific-dash",
                "foo.arch-specific-dash.i386",
            ],
        ),
        (
            "foo",
            "foo.arch-specific-dash",
            ["arch-specific-dash", "foo.arch-specific-dash.i386"],
        ),
    ],
)
def test_packager_provided_files_priority(
    package_name, expected_basename, non_matched
) -> None:
    assert len(non_matched) > 0
    # Inject our custom files
    plugin = initialize_plugin_under_test_preloaded(
        1,
        ppf_test_plugin,
        plugin_name="pff-test-plugin",
    )
    debputy_plugin_feature_set = _fetch_debputy_plugin_feature_set(plugin)
    binary_packages = binary_package_table(faked_binary_package(package_name))
    all_entries_base = [x for x in non_matched]

    for order in (0, len(all_entries_base), None):
        all_entries = all_entries_base.copy()
        print(f"Order: {order}")
        if order is not None:
            all_entries.insert(order, expected_basename)
        else:
            all_entries.append(expected_basename)
            # Ensure there are no order dependencies in the test by randomizing it.
            random.shuffle(all_entries)

        debian_dir = build_virtual_file_system(all_entries)
        ppf = detect_all_packager_provided_files(
            debputy_plugin_feature_set.packager_provided_files,
            debian_dir,
            binary_packages,
        )
        assert package_name in ppf
        all_matched = ppf[package_name].auto_installable
        assert len(all_matched) == 1
        matched = all_matched[0]
        assert expected_basename == matched.path.name


@pytest.mark.parametrize(
    "main_package,secondary_package",
    [
        ("foo", "foo-doc"),
        ("foo.bar", "foo.bar-doc"),
    ],
)
def test_detect_ppf_simple(main_package: str, secondary_package: str) -> None:
    plugin_metadata = plugin_metadata_for_debputys_own_plugin()
    binary_packages = binary_package_table(
        faked_binary_package(main_package, is_main_package=True),
        faked_binary_package(secondary_package),
    )
    debian_dir = build_virtual_file_system(
        [
            "install",
            f"{secondary_package}.install",
            f"{main_package}.docs",
            f"{secondary_package}.docs",
            "copyright",
        ]
    )
    ppf_defs = _ppfs(
        _fake_PPFClassSpec(
            plugin_metadata,
            "install",
        ),
        _fake_PPFClassSpec(
            plugin_metadata,
            "docs",
        ),
        _fake_PPFClassSpec(
            plugin_metadata,
            "copyright",
            packageless_is_fallback_for_all_packages=True,
        ),
    )

    results = detect_all_packager_provided_files(
        ppf_defs,
        debian_dir,
        binary_packages,
    )
    assert main_package in results
    main_matches = results[main_package].auto_installable
    assert {m.path.name for m in main_matches} == {
        "install",
        f"{main_package}.docs",
        "copyright",
    }

    assert secondary_package in results
    sec_matches = results[secondary_package].auto_installable
    expected = {
        f"{secondary_package}.install",
        f"{secondary_package}.docs",
        "copyright",
    }
    assert {m.path.name for m in sec_matches} == expected


def test_detect_ppf_name_segment() -> None:
    plugin_metadata = plugin_metadata_for_debputys_own_plugin()
    binary_packages = binary_package_table(
        faked_binary_package("foo", is_main_package=True),
        faked_binary_package("bar"),
    )
    debian_dir = build_virtual_file_system(
        [
            "fuu.install",
        ]
    )
    ppf_defs = _ppfs(
        _fake_PPFClassSpec(
            plugin_metadata,
            "install",
        ),
    )

    results = detect_all_packager_provided_files(
        ppf_defs,
        debian_dir,
        binary_packages,
    )
    assert "foo" in results
    foo_matches = results["foo"].auto_installable
    assert {m.path.name for m in foo_matches} == {
        "fuu.install",
    }
    foo_install = foo_matches[0]
    assert foo_install.name_segment == "fuu"
    assert foo_install.package_name == "foo"
    assert foo_install.architecture_restriction is None
    assert not foo_install.fuzzy_match

    assert "bar" in results
    assert not results["bar"].auto_installable

    ppf_defs = _ppfs(
        _fake_PPFClassSpec(
            plugin_metadata,
            "install",
            allow_name_segment=False,
        ),
    )

    results = detect_all_packager_provided_files(
        ppf_defs,
        debian_dir,
        binary_packages,
        allow_fuzzy_matches=True,
    )

    assert "foo" in results
    foo_matches = results["foo"].auto_installable
    assert {m.path.name for m in foo_matches} == {"fuu.install"}
    foo_install = foo_matches[0]
    assert foo_install.name_segment == "fuu"
    assert foo_install.package_name == "foo"
    assert foo_install.architecture_restriction is None

    assert "bar" in results
    assert not results["bar"].auto_installable


def test_detect_ppf_fuzzy_match_bug_950723() -> None:
    plugin_metadata = plugin_metadata_for_debputys_own_plugin()
    binary_packages = binary_package_table(
        faked_binary_package("foo", is_main_package=True),
        faked_binary_package("bar"),
    )
    debian_dir = build_virtual_file_system(
        [
            "bar.service",
            "foo@.service",
        ]
    )
    ppf_defs = _ppfs(
        _fake_PPFClassSpec(
            plugin_metadata,
            "service",
            bug_950723=True,
        ),
    )

    results = detect_all_packager_provided_files(
        ppf_defs,
        debian_dir,
        binary_packages,
    )
    assert "foo" in results
    foo_matches = results["foo"].auto_installable
    assert {m.path.name for m in foo_matches} == {"foo@.service"}
    foo_at_service = foo_matches[0]
    # Without bug#950723 AND fuzzy_match, it counts a name segment for the main package.
    assert foo_at_service.name_segment == "foo@"
    assert foo_at_service.package_name == "foo"
    assert foo_at_service.architecture_restriction is None
    assert not foo_at_service.fuzzy_match

    assert "bar" in results
    bar_matches = results["bar"].auto_installable
    assert {m.path.name for m in bar_matches} == {"bar.service"}
    bar_service = bar_matches[0]
    assert bar_service.name_segment is None
    assert bar_service.package_name == "bar"
    assert bar_service.architecture_restriction is None
    assert not bar_service.fuzzy_match

    results = detect_all_packager_provided_files(
        ppf_defs,
        debian_dir,
        binary_packages,
        allow_fuzzy_matches=True,
    )
    assert "foo" in results
    foo_matches = results["foo"].auto_installable
    assert {m.path.name for m in foo_matches} == {"foo@.service"}
    foo_at_service = foo_matches[0]
    assert foo_at_service.name_segment is None
    assert foo_at_service.package_name == "foo"
    assert foo_at_service.architecture_restriction is None

    assert "bar" in results
    bar_matches = results["bar"].auto_installable
    assert {m.path.name for m in bar_matches} == {"bar.service"}
    bar_service = bar_matches[0]
    assert bar_service.name_segment is None
    assert bar_service.package_name == "bar"
    assert bar_service.architecture_restriction is None
    assert not bar_service.fuzzy_match


@requires_levenshtein
def test_detect_ppf_typo() -> None:
    plugin_metadata = plugin_metadata_for_debputys_own_plugin()
    binary_packages = binary_package_table(
        faked_binary_package("foo", is_main_package=True),
        faked_binary_package("bar"),
    )
    debian_dir = build_virtual_file_system(
        [
            "fuu.install",
            "bar.intsall",
        ]
    )
    ppf_defs = _ppfs(
        _fake_PPFClassSpec(
            plugin_metadata,
            "install",
        ),
        _fake_PPFClassSpec(
            plugin_metadata,
            "logcheck.violations.d",
        ),
    )

    results = detect_all_packager_provided_files(
        ppf_defs,
        debian_dir,
        binary_packages,
        detect_typos=True,
    )
    assert "foo" in results
    foo_matches = results["foo"].auto_installable
    assert {m.path.name for m in foo_matches} == {"fuu.install"}
    foo_logcheck = foo_matches[0]
    # Not a typo due to how name segments works with debhelper compat <= 13
    # (but should probably have been one.
    assert foo_logcheck.name_segment == "fuu"
    assert foo_logcheck.package_name == "foo"
    assert foo_logcheck.architecture_restriction is None

    assert "bar" in results
    bar_matches = results["bar"].auto_installable
    assert {m.path.name for m in bar_matches} == {"bar.intsall"}
    bar_logcheck = bar_matches[0]
    assert os.path.basename(bar_logcheck.expected_path) == "bar.install"
    assert bar_logcheck.name_segment is None
    assert bar_logcheck.package_name == "bar"
    assert bar_logcheck.architecture_restriction is None

    debian_dir = build_virtual_file_system(
        [
            # Typo'ed by intention
            "logchcek.violations.d",
            "bar.logchcek.violations.d",
        ]
    )

    results = detect_all_packager_provided_files(
        ppf_defs,
        debian_dir,
        binary_packages,
        detect_typos=True,
    )

    assert "foo" in results
    foo_matches = results["foo"].auto_installable
    assert {m.path.name for m in foo_matches} == {"logchcek.violations.d"}
    foo_logcheck = foo_matches[0]
    assert foo_logcheck.name_segment is None
    assert foo_logcheck.package_name == "foo"
    assert foo_logcheck.architecture_restriction is None
    assert os.path.basename(foo_logcheck.expected_path) == "logcheck.violations.d"

    assert "bar" in results
    bar_matches = results["bar"].auto_installable
    assert {m.path.name for m in bar_matches} == {
        "bar.logchcek.violations.d",
    }
    bar_logcheck = bar_matches[0]
    assert bar_logcheck.name_segment is None
    assert bar_logcheck.package_name == "bar"
    assert bar_logcheck.architecture_restriction is None
    assert os.path.basename(bar_logcheck.expected_path) == "bar.logcheck.violations.d"


def test_debhelper_overlapping_stems() -> None:
    plugin_metadata = plugin_metadata_for_debputys_own_plugin()
    binary_packages = binary_package_table(
        faked_binary_package("foo", is_main_package=True),
    )
    debian_dir = build_virtual_file_system(
        [
            "foo.user.service",
        ]
    )
    ppf_defs = _ppfs(
        _fake_PPFClassSpec(
            plugin_metadata,
            "service",
        ),
        _fake_PPFClassSpec(
            plugin_metadata,
            "user.service",
        ),
    )

    results = detect_all_packager_provided_files(
        ppf_defs,
        debian_dir,
        binary_packages,
    )
    assert "foo" in results
    foo_matches = results["foo"].auto_installable
    assert {m.path.name for m in foo_matches} == {
        "foo.user.service",
    }
    matched_file = foo_matches[0]
    assert matched_file.definition.stem == "user.service"
    assert matched_file.name_segment is None
    assert matched_file.package_name == "foo"
    assert matched_file.architecture_restriction is None

    debian_dir = build_virtual_file_system(
        [
            "foo.named.service",
        ]
    )
    results = detect_all_packager_provided_files(
        ppf_defs,
        debian_dir,
        binary_packages,
    )
    assert "foo" in results
    foo_matches = results["foo"].auto_installable
    assert {m.path.name for m in foo_matches} == {"foo.named.service"}
    matched_file = foo_matches[0]
    assert matched_file.definition.stem == "service"
    assert matched_file.name_segment == "named"
    assert matched_file.package_name == "foo"
    assert matched_file.architecture_restriction is None


@requires_levenshtein
def test_debhelper_overlapping_stems_typo_check() -> None:
    plugin_metadata = plugin_metadata_for_debputys_own_plugin()
    binary_packages = binary_package_table(
        faked_binary_package("foo", is_main_package=True),
    )
    debian_dir = build_virtual_file_system(
        [
            "foo.uxxr.service",
        ]
    )
    ppf_defs = _ppfs(
        _fake_PPFClassSpec(
            plugin_metadata,
            "service",
        ),
        _fake_PPFClassSpec(
            plugin_metadata,
            "user.service",
        ),
    )
    results = detect_all_packager_provided_files(
        ppf_defs,
        debian_dir,
        binary_packages,
        detect_typos=True,
    )
    assert "foo" in results
    foo_matches = results["foo"].auto_installable
    assert {m.path.name for m in foo_matches} == {"foo.uxxr.service"}
    foo_install = foo_matches[0]
    # While it could just as well be a named segment, we assume it is a typo.
    assert foo_install.definition.stem == "user.service"
    assert foo_install.name_segment is None
    assert foo_install.package_name == "foo"
    assert foo_install.architecture_restriction is None
    assert os.path.basename(foo_install.expected_path) == "foo.user.service"


def _fetch_debputy_plugin_feature_set(
    plugin: InitializedPluginUnderTest,
) -> PluginProvidedFeatureSet:
    # Very much not public API, but we need it to avoid testing on production data (also, it is hard to find
    # relevant debputy files for all the cases we want to test).
    return cast("InitializedPluginUnderTestImpl", plugin)._feature_set


def _fake_PPFClassSpec(
    debputy_plugin_metadata: DebputyPluginMetadata,
    stem: str,
    *,
    default_priority: Optional[int] = None,
    packageless_is_fallback_for_all_packages: bool = False,
    bug_950723: bool = False,
    has_active_command: bool = False,
    allow_name_segment: bool = True,
) -> PackagerProvidedFileClassSpec:
    return PackagerProvidedFileClassSpec(
        debputy_plugin_metadata,
        stem,
        "not-a-real-ppf",
        allow_architecture_segment=True,
        allow_name_segment=allow_name_segment,
        default_priority=default_priority,
        default_mode=0o644,
        post_formatting_rewrite=None,
        packageless_is_fallback_for_all_packages=packageless_is_fallback_for_all_packages,
        reservation_only=False,
        formatting_callback=None,
        bug_950723=bug_950723,
        has_active_command=has_active_command,
    )


def _ppfs(
    *ppfs: PackagerProvidedFileClassSpec,
) -> Mapping[str, PackagerProvidedFileClassSpec]:
    return {ppf.stem: ppf for ppf in ppfs}



from typing import (
    TypeVar,
    Optional,
    Union,
    List,
    Callable,
    TYPE_CHECKING,
)

from debputy.manifest_parser.exceptions import ManifestTypeException
from debputy.packages import BinaryPackage
from debputy.util import assume_not_none

if TYPE_CHECKING:
    from debputy.manifest_parser.util import AttributePath
    from debputy.manifest_parser.parser_data import ParserContextData

S = TypeVar("S")
T = TypeVar("T")


def type_mapper_str2package(
    raw_package_name: str,
    ap: "AttributePath",
    opc: Optional["ParserContextData"],
) -> BinaryPackage:
    pc = assume_not_none(opc)
    if "{{" in raw_package_name:
        resolved_package_name = pc.substitution.substitute(raw_package_name, ap.path)
    else:
        resolved_package_name = raw_package_name

    package_name_in_message = raw_package_name
    if resolved_package_name != raw_package_name:
        package_name_in_message = f'"{resolved_package_name}" ["{raw_package_name}"]'

    if not pc.is_known_package(resolved_package_name):
        package_names = ", ".join(pc.binary_packages)
        raise ManifestTypeException(
            f'The value {package_name_in_message} (from "{ap.path}") does not reference a package declared in'
            f" debian/control. Valid options are: {package_names}"
        )
    package_data = pc.binary_package_data(resolved_package_name)
    if package_data.is_auto_generated_package:
        package_names = ", ".join(pc.binary_packages)
        raise ManifestTypeException(
            f'The package name {package_name_in_message} (from "{ap.path}") references an auto-generated package.'
            " However, auto-generated packages are now permitted here. Valid options are:"
            f" {package_names}"
        )
    return package_data.binary_package


def wrap_into_list(
    x: T,
    _ap: "AttributePath",
    _pc: Optional["ParserContextData"],
) -> List[T]:
    return [x]


def normalize_into_list(
    x: Union[T, List[T]],
    _ap: "AttributePath",
    _pc: Optional["ParserContextData"],
) -> List[T]:
    return x if isinstance(x, list) else [x]


def map_each_element(
    mapper: Callable[[S, "AttributePath", Optional["ParserContextData"]], T],
) -> Callable[[List[S], "AttributePath", Optional["ParserContextData"]], List[T]]:
    def _generated_mapper(
        xs: List[S],
        ap: "AttributePath",
        pc: Optional["ParserContextData"],
    ) -> List[T]:
        return [mapper(s, ap[i], pc) for i, s in enumerate(xs)]

    return _generated_mapper



import contextvars
import functools
import inspect
from contextvars import ContextVar
from typing import Optional, Callable, ParamSpec, TypeVar, NoReturn, Union

from debputy.exceptions import (
    UnhandledOrUnexpectedErrorFromPluginError,
    DebputyRuntimeError,
)
from debputy.util import _trace_log, _is_trace_log_enabled

_current_debputy_plugin_cxt_var: ContextVar[Optional[str]] = ContextVar(
    "current_debputy_plugin",
    default=None,
)

P = ParamSpec("P")
R = TypeVar("R")


def current_debputy_plugin_if_present() -> Optional[str]:
    return _current_debputy_plugin_cxt_var.get()


def current_debputy_plugin_required() -> str:
    v = current_debputy_plugin_if_present()
    if v is None:
        raise AssertionError(
            "current_debputy_plugin_required() was called, but no plugin was set."
        )
    return v


def wrap_plugin_code(
    plugin_name: str,
    func: Callable[P, R],
    *,
    non_debputy_exception_handling: Union[bool, Callable[[Exception], NoReturn]] = True,
) -> Callable[P, R]:
    if isinstance(non_debputy_exception_handling, bool):

        runner = run_in_context_of_plugin
        if non_debputy_exception_handling:
            runner = run_in_context_of_plugin_wrap_errors

        def _plugin_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            return runner(plugin_name, func, *args, **kwargs)

        functools.update_wrapper(_plugin_wrapper, func)
        return _plugin_wrapper

    def _wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        try:
            return run_in_context_of_plugin(plugin_name, func, *args, **kwargs)
        except DebputyRuntimeError:
            raise
        except Exception as e:
            non_debputy_exception_handling(e)

    functools.update_wrapper(_wrapper, func)
    return _wrapper


def run_in_context_of_plugin(
    plugin: str,
    func: Callable[P, R],
    *args: P.args,
    **kwargs: P.kwargs,
) -> R:
    context = contextvars.copy_context()
    if _is_trace_log_enabled():
        call_stack = inspect.stack()
        caller: str = "[N/A]"
        for frame in call_stack:
            if frame.filename != __file__:
                try:
                    fname = frame.frame.f_code.co_qualname
                except AttributeError:
                    fname = None
                if fname is None:
                    fname = frame.function
                caller = f"{frame.filename}:{frame.lineno} ({fname})"
                break
        # Do not keep the reference longer than necessary
        del call_stack
        _trace_log(
            f"Switching plugin context to {plugin} at {caller} (from context: {current_debputy_plugin_if_present()})"
        )
    # Wish we could just do a regular set without wrapping it in `context.run`
    context.run(_current_debputy_plugin_cxt_var.set, plugin)
    return context.run(func, *args, **kwargs)


def run_in_context_of_plugin_wrap_errors(
    plugin: str,
    func: Callable[P, R],
    *args: P.args,
    **kwargs: P.kwargs,
) -> R:
    try:
        return run_in_context_of_plugin(plugin, func, *args, **kwargs)
    except DebputyRuntimeError:
        raise
    except Exception as e:
        if plugin != "debputy":
            raise UnhandledOrUnexpectedErrorFromPluginError(
                f"{func.__qualname__} from the plugin {plugin} raised exception that was not expected here."
            ) from e
        else:
            raise AssertionError(
                "Bug in the `debputy` plugin: Unhandled exception."
            ) from e



import dataclasses
from typing import Sequence, Optional, List, Literal, Iterable, Dict, Self

from debputy.manifest_parser.tagging_types import DebputyDispatchableType
from debputy.manifest_parser.util import AttributePath

STD_CONTROL_SCRIPTS = frozenset(
    {
        "preinst",
        "prerm",
        "postinst",
        "postrm",
    }
)
UDEB_CONTROL_SCRIPTS = frozenset(
    {
        "postinst",
        "menutest",
        "isinstallable",
    }
)
ALL_CONTROL_SCRIPTS = STD_CONTROL_SCRIPTS | UDEB_CONTROL_SCRIPTS | {"config"}


@dataclasses.dataclass(slots=True, frozen=True)
class MaintscriptSnippet:
    definition_source: str
    snippet: str
    snippet_order: Optional[Literal["service"]] = None

    def script_content(self) -> str:
        lines = [
            f"# Snippet source: {self.definition_source}\n",
            self.snippet,
        ]
        if not self.snippet.endswith("\n"):
            lines.append("\n")
        return "".join(lines)


class MaintscriptSnippetContainer:
    def __init__(self) -> None:
        self._generic_snippets: List[MaintscriptSnippet] = []
        self._snippets_by_order: Dict[Literal["service"], List[MaintscriptSnippet]] = {}

    def copy(self) -> "MaintscriptSnippetContainer":
        instance = self.__class__()
        instance._generic_snippets = self._generic_snippets.copy()
        instance._snippets_by_order = self._snippets_by_order.copy()
        return instance

    def append(self, maintscript_snippet: MaintscriptSnippet) -> None:
        if maintscript_snippet.snippet_order is None:
            self._generic_snippets.append(maintscript_snippet)
        else:
            if maintscript_snippet.snippet_order not in self._snippets_by_order:
                self._snippets_by_order[maintscript_snippet.snippet_order] = []
            self._snippets_by_order[maintscript_snippet.snippet_order].append(
                maintscript_snippet
            )

    def has_content(self, snippet_order: Optional[Literal["service"]] = None) -> bool:
        if snippet_order is None:
            return bool(self._generic_snippets)
        if snippet_order not in self._snippets_by_order:
            return False
        return bool(self._snippets_by_order[snippet_order])

    def all_snippets(self) -> Iterable[MaintscriptSnippet]:
        yield from self._generic_snippets
        for snippets in self._snippets_by_order.values():
            yield from snippets

    def generate_snippet(
        self,
        tool_with_version: Optional[str] = None,
        snippet_order: Optional[Literal["service"]] = None,
        reverse: bool = False,
    ) -> Optional[str]:
        inner_content = ""
        if snippet_order is None:
            snippets = (
                reversed(self._generic_snippets) if reverse else self._generic_snippets
            )
            inner_content = "".join(s.script_content() for s in snippets)
        elif snippet_order in self._snippets_by_order:
            snippets = self._snippets_by_order[snippet_order]
            if reverse:
                snippets = reversed(snippets)
            inner_content = "".join(s.script_content() for s in snippets)

        if not inner_content:
            return None

        if tool_with_version:
            return (
                f"# Automatically added by {tool_with_version}\n"
                + inner_content
                + "# End automatically added section"
            )
        return inner_content


class DpkgMaintscriptHelperCommand(DebputyDispatchableType):
    __slots__ = ("cmdline", "definition_source")

    def __init__(self, cmdline: Sequence[str], definition_source: str):
        self.cmdline = cmdline
        self.definition_source = definition_source

    @classmethod
    def _finish_cmd(
        cls,
        definition_source: str,
        cmdline: List[str],
        prior_version: Optional[str],
        owning_package: Optional[str],
    ) -> Self:
        if prior_version is not None:
            cmdline.append(prior_version)
        if owning_package is not None:
            if prior_version is None:
                # Empty is allowed according to `man dpkg-maintscript-helper`
                cmdline.append("")
            cmdline.append(owning_package)
        return cls(
            tuple(cmdline),
            definition_source,
        )

    @classmethod
    def rm_conffile(
        cls,
        definition_source: AttributePath,
        conffile: str,
        prior_version: Optional[str] = None,
        owning_package: Optional[str] = None,
    ) -> Self:
        cmdline = ["rm_conffile", conffile]
        return cls._finish_cmd(
            definition_source.path, cmdline, prior_version, owning_package
        )

    @classmethod
    def mv_conffile(
        cls,
        definition_source: AttributePath,
        old_conffile: str,
        new_confile: str,
        prior_version: Optional[str] = None,
        owning_package: Optional[str] = None,
    ) -> Self:
        cmdline = ["mv_conffile", old_conffile, new_confile]
        return cls._finish_cmd(
            definition_source.path, cmdline, prior_version, owning_package
        )

    @classmethod
    def symlink_to_dir(
        cls,
        definition_source: AttributePath,
        pathname: str,
        old_target: str,
        prior_version: Optional[str] = None,
        owning_package: Optional[str] = None,
    ) -> Self:
        cmdline = ["symlink_to_dir", pathname, old_target]
        return cls._finish_cmd(
            definition_source.path, cmdline, prior_version, owning_package
        )

    @classmethod
    def dir_to_symlink(
        cls,
        definition_source: AttributePath,
        pathname: str,
        new_target: str,
        prior_version: Optional[str] = None,
        owning_package: Optional[str] = None,
    ) -> Self:
        cmdline = ["dir_to_symlink", pathname, new_target]
        return cls._finish_cmd(
            definition_source.path, cmdline, prior_version, owning_package
        )



import argparse
import dataclasses
import errno
import logging
import os
from typing import (
    Optional,
    Tuple,
    Mapping,
    FrozenSet,
    Set,
    Union,
    Sequence,
    Iterable,
    Callable,
    Dict,
    TYPE_CHECKING,
    Literal,
)

from debian.debian_support import DpkgArchTable

from debputy._deb_options_profiles import DebBuildOptionsAndProfiles
from debputy.architecture_support import (
    DpkgArchitectureBuildProcessValuesTable,
    dpkg_architecture_table,
)
from debputy.dh.dh_assistant import read_dh_addon_sequences
from debputy.exceptions import DebputyRuntimeError
from debputy.filesystem_scan import FSROOverlay
from debputy.highlevel_manifest import HighLevelManifest
from debputy.highlevel_manifest_parser import YAMLManifestParser
from debputy.integration_detection import determine_debputy_integration_mode
from debputy.packages import (
    SourcePackage,
    BinaryPackage,
    DctrlParser,
)
from debputy.plugin.api import VirtualPath
from debputy.plugin.api.impl import load_plugin_features
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.plugin.api.spec import DebputyIntegrationMode
from debputy.substitution import (
    Substitution,
    VariableContext,
    SubstitutionImpl,
    NULL_SUBSTITUTION,
)
from debputy.util import (
    _error,
    PKGNAME_REGEX,
    resolve_source_date_epoch,
    setup_logging,
    PRINT_COMMAND,
    change_log_level,
    _warn,
    TRACE_LOG,
)

if TYPE_CHECKING:
    from argparse import _SubParsersAction


CommandHandler = Callable[["CommandContext"], None]
ArgparserConfigurator = Callable[[argparse.ArgumentParser], None]


def add_arg(
    *name_or_flags: str,
    **kwargs,
) -> Callable[[argparse.ArgumentParser], None]:
    def _configurator(argparser: argparse.ArgumentParser) -> None:
        argparser.add_argument(
            *name_or_flags,
            **kwargs,
        )

    return _configurator


@dataclasses.dataclass(slots=True, frozen=True)
class CommandArg:
    parsed_args: argparse.Namespace
    plugin_search_dirs: Sequence[str]


@dataclasses.dataclass
class Command:
    handler: Callable[["CommandContext"], None]
    require_substitution: bool = True
    requested_plugins_only: bool = False


def _host_dpo_to_dbo(opt_and_profiles: "DebBuildOptionsAndProfiles", v: str) -> bool:

    if (
        v in opt_and_profiles.deb_build_profiles
        and v not in opt_and_profiles.deb_build_options
    ):
        val = os.environ.get("DEB_BUILD_OPTIONS", "") + " " + v
        _warn(
            f'Copying "{v}" into DEB_BUILD_OPTIONS: It was in DEB_BUILD_PROFILES but not in DEB_BUILD_OPTIONS'
        )
        os.environ["DEB_BUILD_OPTIONS"] = val.lstrip()
        # Note: It will not be immediately visible since `DebBuildOptionsAndProfiles` caches the result
        return True
    return False


class CommandContext:
    def __init__(
        self,
        parsed_args: argparse.Namespace,
        plugin_search_dirs: Sequence[str],
        require_substitution: bool = True,
        requested_plugins_only: bool = False,
    ) -> None:
        self.parsed_args = parsed_args
        self.plugin_search_dirs = plugin_search_dirs
        self._require_substitution = require_substitution
        self._requested_plugins_only = requested_plugins_only
        self._debputy_plugin_feature_set: PluginProvidedFeatureSet = (
            PluginProvidedFeatureSet()
        )
        self._debian_dir = FSROOverlay.create_root_dir("debian", "debian")
        self._mtime: Optional[int] = None
        self._source_variables: Optional[Mapping[str, str]] = None
        self._substitution: Optional[Substitution] = None
        self._requested_plugins: Optional[Sequence[str]] = None
        self._plugins_loaded = False
        self._dctrl_parser: Optional[DctrlParser] = None
        self.debputy_integration_mode: Optional[DebputyIntegrationMode] = None
        self._dctrl_data: Optional[
            Tuple[
                "SourcePackage",
                Mapping[str, "BinaryPackage"],
            ]
        ] = None
        self._package_set: Literal["both", "arch", "indep"] = "both"

    @property
    def package_set(self) -> Literal["both", "arch", "indep"]:
        return self._package_set

    @package_set.setter
    def package_set(self, new_value: Literal["both", "arch", "indep"]) -> None:
        if self._dctrl_parser is not None:
            raise TypeError(
                "package_set cannot be redefined once the debian/control parser has been initialized"
            )
        self._package_set = new_value

    @property
    def debian_dir(self) -> VirtualPath:
        return self._debian_dir

    @property
    def mtime(self) -> int:
        if self._mtime is None:
            self._mtime = resolve_source_date_epoch(
                None,
                substitution=self.substitution,
            )
        return self._mtime

    @property
    def dctrl_parser(self) -> DctrlParser:
        parser = self._dctrl_parser
        if parser is None:
            packages: Union[Set[str], FrozenSet[str]] = frozenset()
            if hasattr(self.parsed_args, "packages"):
                packages = self.parsed_args.packages

            instance = DebBuildOptionsAndProfiles(environ=os.environ)

            dirty = _host_dpo_to_dbo(instance, "nodoc")
            dirty = _host_dpo_to_dbo(instance, "nocheck") or dirty

            if dirty:
                instance = DebBuildOptionsAndProfiles(environ=os.environ)

            parser = DctrlParser(
                packages,  # -p/--package
                set(),  # -N/--no-package
                # binary-indep and binary-indep (dpkg BuildDriver integration only)
                self._package_set == "indep",
                self._package_set == "arch",
                deb_options_and_profiles=instance,
                dpkg_architecture_variables=dpkg_architecture_table(),
                dpkg_arch_query_table=DpkgArchTable.load_arch_table(),
            )
            self._dctrl_parser = parser
        return parser

    def source_package(self) -> SourcePackage:
        source, _ = self._parse_dctrl()
        return source

    def binary_packages(self) -> Mapping[str, "BinaryPackage"]:
        _, binary_package_table = self._parse_dctrl()
        return binary_package_table

    def dpkg_architecture_variables(self) -> DpkgArchitectureBuildProcessValuesTable:
        return self.dctrl_parser.dpkg_architecture_variables

    def requested_plugins(self) -> Sequence[str]:
        if self._requested_plugins is None:
            self._requested_plugins = self._resolve_requested_plugins()
        return self._requested_plugins

    def required_plugins(self) -> Set[str]:
        return set(getattr(self.parsed_args, "required_plugins") or [])

    @property
    def deb_build_options_and_profiles(self) -> "DebBuildOptionsAndProfiles":
        return self.dctrl_parser.deb_options_and_profiles

    @property
    def deb_build_options(self) -> Mapping[str, Optional[str]]:
        return self.deb_build_options_and_profiles.deb_build_options

    def _create_substitution(
        self,
        parsed_args: argparse.Namespace,
        plugin_feature_set: PluginProvidedFeatureSet,
        debian_dir: VirtualPath,
    ) -> Substitution:
        requested_subst = self._require_substitution
        if hasattr(parsed_args, "substitution"):
            requested_subst = parsed_args.substitution
        if requested_subst is False and self._require_substitution:
            _error(f"--no-substitution cannot be used with {parsed_args.command}")
        if self._require_substitution or requested_subst is not False:
            variable_context = VariableContext(debian_dir)
            return SubstitutionImpl(
                plugin_feature_set=plugin_feature_set,
                unresolvable_substitutions=frozenset(["PACKAGE"]),
                variable_context=variable_context,
            )
        return NULL_SUBSTITUTION

    def load_plugins(self) -> PluginProvidedFeatureSet:
        if not self._plugins_loaded:
            requested_plugins = None
            required_plugins = self.required_plugins()
            if self._requested_plugins_only:
                requested_plugins = self.requested_plugins()
            debug_mode = getattr(self.parsed_args, "debug_mode", False)
            load_plugin_features(
                self.plugin_search_dirs,
                self.substitution,
                requested_plugins_only=requested_plugins,
                required_plugins=required_plugins,
                plugin_feature_set=self._debputy_plugin_feature_set,
                debug_mode=debug_mode,
            )
            self._plugins_loaded = True
        return self._debputy_plugin_feature_set

    @staticmethod
    def _plugin_from_dependency_field(dep_field: str) -> Iterable[str]:
        package_prefix = "debputy-plugin-"
        for dep_clause in (d.strip() for d in dep_field.split(",")):
            dep = dep_clause.split("|")[0].strip()
            if not dep.startswith(package_prefix):
                continue
            m = PKGNAME_REGEX.search(dep)
            assert m
            package_name = m.group(0)
            plugin_name = package_name[len(package_prefix) :]
            yield plugin_name

    def _resolve_requested_plugins(self) -> Sequence[str]:
        source_package, _ = self._parse_dctrl()
        bd = source_package.fields.get("Build-Depends", "")
        plugins = list(self._plugin_from_dependency_field(bd))
        for field_name in ("Build-Depends-Arch", "Build-Depends-Indep"):
            f = source_package.fields.get(field_name)
            if not f:
                continue
            for plugin in self._plugin_from_dependency_field(f):
                raise DebputyRuntimeError(
                    f"Cannot load plugins via {field_name}:"
                    f" Please move debputy-plugin-{plugin} dependency to Build-Depends."
                )

        return plugins

    @property
    def substitution(self) -> Substitution:
        if self._substitution is None:
            self._substitution = self._create_substitution(
                self.parsed_args,
                self._debputy_plugin_feature_set,
                self.debian_dir,
            )
        return self._substitution

    def must_be_called_in_source_root(self) -> None:
        if self.debian_dir.get("control") is None:
            _error(
                "This subcommand must be run from a source package root; expecting debian/control to exist."
            )

    def _parse_dctrl(
        self,
    ) -> Tuple[
        "SourcePackage",
        Mapping[str, "BinaryPackage"],
    ]:
        if self._dctrl_data is None:
            try:
                debian_control = self.debian_dir.get("control")
                if debian_control is None:
                    raise FileNotFoundError(
                        errno.ENOENT,
                        os.strerror(errno.ENOENT),
                        os.path.join(self.debian_dir.fs_path, "control"),
                    )
                with debian_control.open() as fd:
                    _, source_package, binary_packages = (
                        self.dctrl_parser.parse_source_debian_control(
                            fd,
                        )
                    )
            except FileNotFoundError:
                # We are not using `must_be_called_in_source_root`, because we (in this case) require
                # the file to be readable (that is, parse_source_debian_control can also raise a
                # FileNotFoundError when trying to open the file).
                _error(
                    "This subcommand must be run from a source package root; expecting debian/control to exist."
                )

            self._dctrl_data = (
                source_package,
                binary_packages,
            )

        return self._dctrl_data

    @property
    def has_dctrl_file(self) -> bool:
        debian_control = self.debian_dir.get("control")
        return debian_control is not None

    def resolve_integration_mode(
        self,
        require_integration: bool = True,
    ) -> DebputyIntegrationMode:
        integration_mode = self.debputy_integration_mode
        if integration_mode is None:
            r = read_dh_addon_sequences(self.debian_dir)
            bd_sequences, dr_sequences, _ = r
            all_sequences = bd_sequences | dr_sequences
            integration_mode = determine_debputy_integration_mode(
                self.source_package().fields,
                all_sequences,
            )
            if integration_mode is None and not require_integration:
                _error(
                    "Cannot resolve the integration mode expected for this package. Is this package using `debputy`?"
                )
            self.debputy_integration_mode = integration_mode
        return integration_mode

    def set_log_level_for_build_subcommand(self) -> Optional[int]:
        parsed_args = self.parsed_args
        log_level: Optional[int] = None
        if os.environ.get("DH_VERBOSE", "") != "":
            log_level = PRINT_COMMAND
        if parsed_args.debug_mode:
            log_level = logging.DEBUG
        if log_level is not None:
            change_log_level(log_level)
        return log_level

    def manifest_parser(
        self,
        *,
        manifest_path: Optional[str] = None,
    ) -> YAMLManifestParser:
        substitution = self.substitution
        dctrl_parser = self.dctrl_parser

        source_package, binary_packages = self._parse_dctrl()

        if self.parsed_args.debputy_manifest is not None:
            manifest_path = self.parsed_args.debputy_manifest
        if manifest_path is None:
            manifest_path = os.path.join(self.debian_dir.fs_path, "debputy.manifest")
        return YAMLManifestParser(
            manifest_path,
            source_package,
            binary_packages,
            substitution,
            dctrl_parser.dpkg_architecture_variables,
            dctrl_parser.dpkg_arch_query_table,
            dctrl_parser.deb_options_and_profiles,
            self.load_plugins(),
            self.resolve_integration_mode(),
            debian_dir=self.debian_dir,
        )

    def parse_manifest(
        self,
        *,
        manifest_path: Optional[str] = None,
    ) -> HighLevelManifest:
        substitution = self.substitution
        manifest_required = False

        if self.parsed_args.debputy_manifest is not None:
            manifest_path = self.parsed_args.debputy_manifest
            manifest_required = True
        if manifest_path is None:
            manifest_path = os.path.join(self.debian_dir.fs_path, "debputy.manifest")
        parser = self.manifest_parser(manifest_path=manifest_path)

        os.environ["SOURCE_DATE_EPOCH"] = substitution.substitute(
            "{{SOURCE_DATE_EPOCH}}",
            "Internal resolution",
        )
        if os.path.isfile(manifest_path):
            return parser.parse_manifest()
        if manifest_required:
            _error(f'The path "{manifest_path}" is not a file!')
        return parser.build_manifest()


class CommandBase:
    __slots__ = ()

    def configure(self, argparser: argparse.ArgumentParser) -> None:
        # Does nothing by default
        pass

    def __call__(self, command_arg: CommandArg) -> None:
        raise NotImplementedError


class SubcommandBase(CommandBase):
    __slots__ = ("name", "aliases", "help_description")

    def __init__(
        self,
        name: str,
        *,
        aliases: Sequence[str] = tuple(),
        help_description: Optional[str] = None,
    ) -> None:
        self.name = name
        self.aliases = aliases
        self.help_description = help_description

    def add_subcommand_to_subparser(
        self,
        subparser: "_SubParsersAction",
    ) -> argparse.ArgumentParser:
        parser = subparser.add_parser(
            self.name,
            aliases=self.aliases,
            help=self.help_description,
            allow_abbrev=False,
        )
        self.configure(parser)
        return parser


class GenericSubCommand(SubcommandBase):
    __slots__ = (
        "_handler",
        "_configure_handler",
        "_require_substitution",
        "_requested_plugins_only",
        "_log_only_to_stderr",
        "_default_log_level",
    )

    def __init__(
        self,
        name: str,
        handler: Callable[[CommandContext], None],
        *,
        aliases: Sequence[str] = tuple(),
        help_description: Optional[str] = None,
        configure_handler: Optional[Callable[[argparse.ArgumentParser], None]] = None,
        require_substitution: bool = True,
        requested_plugins_only: bool = False,
        log_only_to_stderr: bool = False,
        default_log_level: Union[int, Callable[[CommandContext], int]] = logging.INFO,
    ) -> None:
        super().__init__(name, aliases=aliases, help_description=help_description)
        self._handler = handler
        self._configure_handler = configure_handler
        self._require_substitution = require_substitution
        self._requested_plugins_only = requested_plugins_only
        self._log_only_to_stderr = log_only_to_stderr
        self._default_log_level = default_log_level

    def configure_handler(
        self,
        handler: Callable[[argparse.ArgumentParser], None],
    ) -> None:
        if self._configure_handler is not None:
            raise TypeError("Only one argument handler can be provided")
        self._configure_handler = handler

    def configure(self, argparser: argparse.ArgumentParser) -> None:
        handler = self._configure_handler
        if handler is not None:
            handler(argparser)

    def __call__(self, command_arg: CommandArg) -> None:
        context = CommandContext(
            command_arg.parsed_args,
            command_arg.plugin_search_dirs,
            self._require_substitution,
            self._requested_plugins_only,
        )
        if self._log_only_to_stderr:
            setup_logging(reconfigure_logging=True, log_only_to_stderr=True)

        default_log_level = self._default_log_level
        if isinstance(default_log_level, int):
            level = default_log_level
        else:
            assert callable(default_log_level)
            level = default_log_level(context)
        change_log_level(level)
        if level > logging.DEBUG and context.parsed_args.debug_mode:
            change_log_level(logging.DEBUG)
        if level > TRACE_LOG and os.environ.get("DEBPUTY_DEBUG", "") == "trace":
            change_log_level(TRACE_LOG)
        return self._handler(context)


class DispatchingCommandMixin(CommandBase):
    __slots__ = ()

    def add_subcommand(self, subcommand: SubcommandBase) -> None:
        raise NotImplementedError

    def add_dispatching_subcommand(
        self,
        name: str,
        dest: str,
        *,
        aliases: Sequence[str] = tuple(),
        help_description: Optional[str] = None,
        metavar: str = "command",
        default_subcommand: Optional[str] = None,
    ) -> "DispatcherCommand":
        ds = DispatcherCommand(
            name,
            dest,
            aliases=aliases,
            help_description=help_description,
            metavar=metavar,
            default_subcommand=default_subcommand,
        )
        self.add_subcommand(ds)
        return ds

    def register_subcommand(
        self,
        name: Union[str, Sequence[str]],
        *,
        help_description: Optional[str] = None,
        argparser: Optional[
            Union[ArgparserConfigurator, Sequence[ArgparserConfigurator]]
        ] = None,
        require_substitution: bool = True,
        requested_plugins_only: bool = False,
        log_only_to_stderr: bool = False,
        default_log_level: Union[int, Callable[[CommandContext], int]] = logging.INFO,
    ) -> Callable[[CommandHandler], GenericSubCommand]:
        if isinstance(name, str):
            cmd_name = name
            aliases = []
        else:
            cmd_name = name[0]
            aliases = name[1:]

        if argparser is not None and not callable(argparser):
            args = argparser

            def _wrapper(parser: argparse.ArgumentParser) -> None:
                for configurator in args:
                    configurator(parser)

            argparser = _wrapper

        def _annotation_impl(func: CommandHandler) -> GenericSubCommand:
            subcommand = GenericSubCommand(
                cmd_name,
                func,
                aliases=aliases,
                help_description=help_description,
                require_substitution=require_substitution,
                requested_plugins_only=requested_plugins_only,
                log_only_to_stderr=log_only_to_stderr,
                default_log_level=default_log_level,
            )
            self.add_subcommand(subcommand)
            if argparser is not None:
                subcommand.configure_handler(argparser)

            return subcommand

        return _annotation_impl


class DispatcherCommand(SubcommandBase, DispatchingCommandMixin):
    __slots__ = (
        "_subcommands",
        "_aliases",
        "_dest",
        "_metavar",
        "_required",
        "_default_subcommand",
        "_argparser",
    )

    def __init__(
        self,
        name: str,
        dest: str,
        *,
        aliases: Sequence[str] = tuple(),
        help_description: Optional[str] = None,
        metavar: str = "command",
        default_subcommand: Optional[str] = None,
    ) -> None:
        super().__init__(name, aliases=aliases, help_description=help_description)
        self._aliases: Dict[str, SubcommandBase] = {}
        self._subcommands: Dict[str, SubcommandBase] = {}
        self._dest = dest
        self._metavar = metavar
        self._default_subcommand = default_subcommand
        self._argparser: Optional[argparse.ArgumentParser] = None

    def add_subcommand(self, subcommand: SubcommandBase) -> None:
        all_names = [subcommand.name]
        if subcommand.aliases:
            all_names.extend(subcommand.aliases)
        aliases = self._aliases
        for n in all_names:
            if n in aliases:
                raise ValueError(
                    f"Internal error: Multiple handlers for {n} on topic {self.name}"
                )

            aliases[n] = subcommand
        self._subcommands[subcommand.name] = subcommand

    def configure(self, argparser: argparse.ArgumentParser) -> None:
        if self._argparser is not None:
            raise TypeError("Cannot configure twice!")
        self._argparser = argparser
        subcommands = self._subcommands
        if not subcommands:
            raise ValueError(
                f"Internal error: No subcommands for subcommand {self.name} (then why do we have it?)"
            )
        default_subcommand = self._default_subcommand
        required = default_subcommand is None
        if (
            default_subcommand is not None
            and default_subcommand not in ("--help", "-h")
            and default_subcommand not in subcommands
        ):
            raise ValueError(
                f"Internal error: Subcommand {self.name} should have {default_subcommand} as default,"
                " but it was not registered?"
            )
        subparser = argparser.add_subparsers(
            dest=self._dest,
            required=required,
            metavar=self._metavar,
        )
        for subcommand in subcommands.values():
            subcommand.add_subcommand_to_subparser(subparser)

    def has_command(self, command: str) -> bool:
        return command in self._aliases

    def __call__(self, command_arg: CommandArg) -> None:
        argparser = self._argparser
        assert argparser is not None
        v = getattr(command_arg.parsed_args, self._dest, None)
        if v is None:
            v = self._default_subcommand
            if v in ("--help", "-h"):
                argparser.parse_args([v])
                _error("Missing command", prog=argparser.prog)

        assert (
            v is not None
        ), f"Internal error: No default subcommand and argparse did not provide the required subcommand {self._dest}?"
        assert (
            v in self._aliases
        ), f"Internal error: {v} was accepted as a topic, but it was not registered?"
        self._aliases[v](command_arg)


ROOT_COMMAND = DispatcherCommand(
    "root",
    dest="command",
    metavar="COMMAND",
)



from typing import Optional, Callable

__version__ = "N/A"

IS_RELEASE_BUILD = False
_RUN_FROM_SOURCE = False

if __version__ in ("N/A",):
    _RUN_FROM_SOURCE = True
    import subprocess

    class LazyString:
        def __init__(self, initializer: Callable[[], str]) -> None:
            self._initializer = initializer
            self._value: Optional[str] = None

        def __str__(self) -> str:
            value = object.__getattribute__(self, "_value")
            if value is None:
                value = object.__getattribute__(self, "_initializer")()
                object.__setattr__(self, "_value", value)
            return value

        def __getattribute__(self, item):
            value = str(self)
            return getattr(value, item)

        def __contains__(self, item):
            return item in str(self)

    def _initialize_version() -> str:
        try:
            devnull: Optional[int] = subprocess.DEVNULL
        except AttributeError:
            devnull = None  # Not supported, but not critical

        try:
            v = (
                subprocess.check_output(
                    ["git", "describe", "--tags"],
                    stderr=devnull,
                )
                .strip()
                .decode("utf-8")
            )
        except (subprocess.CalledProcessError, FileNotFoundError):
            try:
                v = (
                    subprocess.check_output(
                        ["dpkg-parsechangelog", "-SVersion"],
                        stderr=devnull,
                    )
                    .strip()
                    .decode("utf-8")
                )

            except (subprocess.CalledProcessError, FileNotFoundError):
                v = "N/A"

        if v.startswith("archive/"):
            v = v[8:]

        if v.startswith("debian/"):
            v = v[7:]
        return v

    __version__ = LazyString(_initialize_version)
    IS_RELEASE_BUILD = False

else:
    # Disregard snapshot versions (gbp dch -S) as "release builds"
    IS_RELEASE_BUILD = ".gbp" not in __version__



import dataclasses
from typing import Dict, List, Tuple, Sequence, Any, Optional, Type

from debputy.manifest_parser.declarative_parser import ParserGenerator
from debputy.plugin.api.impl_types import (
    DebputyPluginMetadata,
    PackagerProvidedFileClassSpec,
    MetadataOrMaintscriptDetector,
    PluginProvidedManifestVariable,
    PluginProvidedPackageProcessor,
    PluginProvidedDiscardRule,
    ServiceManagerDetails,
    PluginProvidedKnownPackagingFile,
    PluginProvidedTypeMapping,
    PluginProvidedBuildSystemAutoDetection,
)
from debputy.plugin.api.parser_tables import (
    SUPPORTED_DISPATCHABLE_OBJECT_PARSERS,
    SUPPORTED_DISPATCHABLE_TABLE_PARSERS,
)
from debputy.plugin.debputy.to_be_api_types import BuildSystemRule


def _initialize_parser_generator() -> ParserGenerator:
    pg = ParserGenerator()

    for path, ref_doc in SUPPORTED_DISPATCHABLE_OBJECT_PARSERS.items():
        pg.add_object_parser(path, parser_documentation=ref_doc)

    for rt, path in SUPPORTED_DISPATCHABLE_TABLE_PARSERS.items():
        pg.add_table_parser(rt, path)

    return pg


@dataclasses.dataclass(slots=True)
class PluginProvidedFeatureSet:
    plugin_data: Dict[str, DebputyPluginMetadata] = dataclasses.field(
        default_factory=dict
    )
    packager_provided_files: Dict[str, PackagerProvidedFileClassSpec] = (
        dataclasses.field(default_factory=dict)
    )
    metadata_maintscript_detectors: Dict[str, List[MetadataOrMaintscriptDetector]] = (
        dataclasses.field(default_factory=dict)
    )
    manifest_variables: Dict[str, PluginProvidedManifestVariable] = dataclasses.field(
        default_factory=dict
    )
    all_package_processors: Dict[Tuple[str, str], PluginProvidedPackageProcessor] = (
        dataclasses.field(default_factory=dict)
    )
    auto_discard_rules: Dict[str, PluginProvidedDiscardRule] = dataclasses.field(
        default_factory=dict
    )
    service_managers: Dict[str, ServiceManagerDetails] = dataclasses.field(
        default_factory=dict
    )
    known_packaging_files: Dict[str, PluginProvidedKnownPackagingFile] = (
        dataclasses.field(default_factory=dict)
    )
    mapped_types: Dict[Any, PluginProvidedTypeMapping] = dataclasses.field(
        default_factory=dict
    )
    manifest_parser_generator: ParserGenerator = dataclasses.field(
        default_factory=_initialize_parser_generator
    )
    auto_detectable_build_systems: Dict[
        Type[BuildSystemRule], PluginProvidedBuildSystemAutoDetection
    ] = dataclasses.field(default_factory=dict)

    def package_processors_in_order(self) -> Sequence[PluginProvidedPackageProcessor]:
        order = []
        delayed = []
        for plugin_processor in self.all_package_processors.values():
            if not plugin_processor.dependencies:
                order.append(plugin_processor)
            else:
                delayed.append(plugin_processor)

        # At the time of writing, insert order will work as a plugin cannot declare
        # dependencies out of order in the current version.  However, we want to
        # ensure dependencies are taken a bit seriously, so we ensure that processors
        # without dependencies are run first.  This should weed out anything that
        # needs dependencies but do not add them.
        #
        # It is still far from as any dependency issues will be hidden if you just
        # add a single dependency.
        order.extend(delayed)
        return order



from debputy.plugin.api.test_api.test_impl import (
    package_metadata_context,
    initialize_plugin_under_test,
    manifest_variable_resolution_context,
)
from debputy.plugin.api.test_api.test_spec import (
    RegisteredPackagerProvidedFile,
    build_virtual_file_system,
    InitializedPluginUnderTest,
    DEBPUTY_TEST_AGAINST_INSTALLED_PLUGINS,
)

__all__ = [
    "initialize_plugin_under_test",
    "RegisteredPackagerProvidedFile",
    "build_virtual_file_system",
    "InitializedPluginUnderTest",
    "package_metadata_context",
    "manifest_variable_resolution_context",
    "DEBPUTY_TEST_AGAINST_INSTALLED_PLUGINS",
]



import functools
import itertools
import os
import re
import subprocess
from typing import (
    Union,
    Sequence,
    Optional,
    Iterable,
    List,
    Iterator,
    Tuple,
    FrozenSet,
)

from debputy.dh.dh_assistant import (
    resolve_active_and_inactive_dh_commands,
    DhListCommands,
)
from debputy.linting.lint_util import LintState
from debputy.lsp.debputy_ls import DebputyLanguageServer
from debputy.lsp.lsp_features import (
    lint_diagnostics,
    lsp_standard_handler,
    lsp_completer,
    SecondaryLanguage,
    LanguageDispatchRule,
)
from debputy.lsp.quickfixes import propose_correct_text_quick_fix
from debputy.lsp.spellchecking import spellcheck_line
from debputy.lsp.text_util import (
    LintCapablePositionCodec,
)
from debputy.lsprotocol.types import (
    CompletionItem,
    Diagnostic,
    CompletionList,
    CompletionParams,
    TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL,
    TEXT_DOCUMENT_CODE_ACTION,
)
from debputy.util import detect_possible_typo

try:
    from debputy.lsp.vendoring._deb822_repro.locatable import (
        Position as TEPosition,
        Range as TERange,
    )

    from pygls.server import LanguageServer
    from pygls.workspace import TextDocument
except ImportError:
    pass


_CONTAINS_TAB_OR_COLON = re.compile(r"[\t:]")
_WORDS_RE = re.compile("([a-zA-Z0-9_-]+)")
_MAKE_ERROR_RE = re.compile(r"^[^:]+:(\d+):\s*(\S.+)")


_KNOWN_TARGETS = {
    "binary",
    "binary-arch",
    "binary-indep",
    "build",
    "build-arch",
    "build-indep",
    "clean",
}

_COMMAND_WORDS = frozenset(
    {
        "export",
        "ifeq",
        "ifneq",
        "ifdef",
        "ifndef",
        "endif",
        "else",
    }
)
_DISPATCH_RULE = LanguageDispatchRule.new_rule(
    "debian/rules",
    "debian/rules",
    [
        # emacs's name (there is no debian-rules mode)
        SecondaryLanguage("makefile-gmake", filename_based_lookup=True),
        # vim's name (there is no debrules and it does not use the official makefile language name)
        SecondaryLanguage("make", filename_based_lookup=True),
        # LSP's official language ID for Makefile
        SecondaryLanguage("makefile", filename_based_lookup=True),
    ],
)


def _as_hook_targets(command_name: str) -> Iterable[str]:
    for prefix, suffix in itertools.product(
        ["override_", "execute_before_", "execute_after_"],
        ["", "-arch", "-indep"],
    ):
        yield f"{prefix}{command_name}{suffix}"


lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_CODE_ACTION)
lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)


@functools.lru_cache
def _is_project_trusted(source_root: str) -> bool:
    return os.environ.get("DEBPUTY_TRUST_PROJECT", "0") == "1"


def _run_make_dryrun(
    lint_state: LintState,
    source_root: str,
    lines: List[str],
) -> None:
    if not _is_project_trusted(source_root):
        return None
    try:
        make_res = subprocess.run(
            ["make", "--dry-run", "-f", "-", "debhelper-fail-me"],
            input="".join(lines).encode("utf-8"),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            cwd=source_root,
            timeout=1,
        )
    except (FileNotFoundError, subprocess.TimeoutExpired):
        pass
    else:
        if make_res.returncode != 0:
            make_output = make_res.stderr.decode("utf-8")
            m = _MAKE_ERROR_RE.match(make_output)
            if m:
                # We want it zero-based and make reports it one-based
                line_of_error = int(m.group(1)) - 1
                msg = m.group(2).strip()
                error_range = TERange(
                    TEPosition(
                        line_of_error,
                        0,
                    ),
                    TEPosition(
                        line_of_error + 1,
                        0,
                    ),
                )
                lint_state.emit_diagnostic(
                    error_range,
                    f"make error: {msg}",
                    "error",
                    "debputy",
                )
    return


def iter_make_lines(
    lint_state: LintState,
    lines: List[str],
) -> Iterator[Tuple[int, str]]:
    skip_next_line = False
    is_extended_comment = False
    for line_no, line in enumerate(lines):
        skip_this = skip_next_line
        skip_next_line = False
        if line.rstrip().endswith("\\"):
            skip_next_line = True

        if skip_this:
            if is_extended_comment:
                spellcheck_line(lint_state, line_no, line)
            continue

        if line.startswith("#"):
            spellcheck_line(lint_state, line_no, line)
            is_extended_comment = skip_next_line
            continue
        is_extended_comment = False

        if line.startswith("\t") or line.isspace():
            continue

        is_extended_comment = False
        # We are not really dealing with extension lines at the moment (other than for spellchecking),
        # since nothing needs it
        yield line_no, line


def _forbidden_hook_targets(dh_commands: DhListCommands) -> FrozenSet[str]:
    if not dh_commands.disabled_commands:
        return frozenset()
    return frozenset(
        itertools.chain.from_iterable(
            _as_hook_targets(c) for c in dh_commands.disabled_commands
        )
    )


@lint_diagnostics(_DISPATCH_RULE)
async def _lint_debian_rules(lint_state: LintState) -> None:
    lines = lint_state.lines
    path = lint_state.path
    source_root = os.path.dirname(os.path.dirname(path))
    if source_root == "":
        source_root = "."

    _run_make_dryrun(lint_state, source_root, lines)
    dh_sequencer_data = lint_state.dh_sequencer_data
    dh_sequences = dh_sequencer_data.sequences
    dh_commands = resolve_active_and_inactive_dh_commands(
        dh_sequences,
        source_root=source_root,
    )
    if dh_commands.active_commands:
        all_hook_targets = {
            ht for c in dh_commands.active_commands for ht in _as_hook_targets(c)
        }
        all_hook_targets.update(_KNOWN_TARGETS)
    else:
        all_hook_targets = _KNOWN_TARGETS

    missing_targets = {}
    forbidden_hook_targets = _forbidden_hook_targets(dh_commands)
    all_allowed_hook_targets = all_hook_targets - forbidden_hook_targets

    for line_no, line in iter_make_lines(lint_state, lines):
        try:
            colon_idx = line.index(":")
            if len(line) > colon_idx + 1 and line[colon_idx + 1] == "=":
                continue
        except ValueError:
            continue
        target_substring = line[0:colon_idx]
        if "=" in target_substring or "$(for" in target_substring:
            continue
        for i, m in enumerate(_WORDS_RE.finditer(target_substring)):
            target = m.group(1)
            if i == 0 and (target in _COMMAND_WORDS or target.startswith("(")):
                break
            if "%" in target or "$" in target:
                continue
            if target in forbidden_hook_targets:
                pos, endpos = m.span(1)
                r = TERange(
                    TEPosition(
                        line_no,
                        pos,
                    ),
                    TEPosition(
                        line_no,
                        endpos,
                    ),
                )
                lint_state.emit_diagnostic(
                    r,
                    f"The hook target {target} will not be run due to the choice of sequences.",
                    "error",
                    "debputy",
                )
                continue

            if target in all_allowed_hook_targets or target in missing_targets:
                continue
            pos, endpos = m.span(1)
            hook_location = line_no, pos, endpos
            missing_targets[target] = hook_location

    for target, (line_no, pos, endpos) in missing_targets.items():
        candidates = detect_possible_typo(target, all_allowed_hook_targets)
        if not candidates and not target.startswith(
            ("override_", "execute_before_", "execute_after_")
        ):
            continue

        r = TERange(
            TEPosition(
                line_no,
                pos,
            ),
            TEPosition(
                line_no,
                endpos,
            ),
        )
        if candidates:
            msg = f"Target {target} looks like a typo of a known target"
        else:
            msg = f"Unknown rules dh hook target {target}"
        if candidates:
            fixes = [propose_correct_text_quick_fix(c) for c in candidates]
        else:
            fixes = []
        lint_state.emit_diagnostic(
            r,
            msg,
            "warning",
            "debputy",
            quickfixes=fixes,
        )


@lsp_completer(_DISPATCH_RULE)
def _debian_rules_completions(
    ls: "DebputyLanguageServer",
    params: CompletionParams,
) -> Optional[Union[CompletionList, Sequence[CompletionItem]]]:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lines = doc.lines
    server_position = doc.position_codec.position_from_client_units(
        lines, params.position
    )

    line = lines[server_position.line]
    line_start = line[0 : server_position.character]

    if _CONTAINS_TAB_OR_COLON.search(line_start):
        return None

    source_root = os.path.dirname(os.path.dirname(doc.path))
    dh_sequencer_data = ls.lint_state(doc).dh_sequencer_data
    dh_sequences = dh_sequencer_data.sequences
    dh_commands = resolve_active_and_inactive_dh_commands(
        dh_sequences,
        source_root=source_root,
    )
    if not dh_commands.active_commands:
        return None
    items = [
        CompletionItem(ht)
        for c in dh_commands.active_commands
        for ht in _as_hook_targets(c)
    ]

    return items



from typing import (
    Optional,
    TYPE_CHECKING,
)

from debputy.filesystem_scan import FSROOverlay
from debputy.manifest_parser.base_types import BuildEnvironmentDefinition
from debputy.manifest_parser.util import AttributePath
from debputy.plugin.debputy.to_be_api_types import (
    BuildSystemRule,
)
from debputy.plugin.plugin_state import run_in_context_of_plugin_wrap_errors
from debputy.util import (
    _error,
    _debug_log,
)

if TYPE_CHECKING:
    from debputy.highlevel_manifest import HighLevelManifest


def default_build_environment_only(
    manifest: "HighLevelManifest",
) -> BuildEnvironmentDefinition:
    build_envs = manifest.build_environments
    if build_envs.environments:
        _error(
            'When automatic build system detection is used, the manifest cannot use "build-environments"'
        )
    build_env = build_envs.default_environment
    assert build_env is not None
    return build_env


def auto_detect_buildsystem(
    manifest: "HighLevelManifest",
) -> Optional[BuildSystemRule]:
    auto_detectable_build_systems = (
        manifest.plugin_provided_feature_set.auto_detectable_build_systems
    )
    excludes = set()
    options = []
    _debug_log("Auto-detecting build systems.")
    source_root = FSROOverlay.create_root_dir("", ".")
    for ppadbs in auto_detectable_build_systems.values():
        detected = ppadbs.detector(source_root)
        if not isinstance(detected, bool):
            _error(
                f'The auto-detector for the build system {ppadbs.manifest_keyword} returned a "non-bool"'
                f" ({detected!r}), which could be a bug in the plugin or the plugin relying on a newer"
                " version of `debputy` that changed the auto-detection protocol."
            )
        if not detected:
            _debug_log(
                f"Skipping build system {ppadbs.manifest_keyword}: Detector returned False!"
            )
            continue
        _debug_log(
            f"Considering build system {ppadbs.manifest_keyword} as its Detector returned True!"
        )
        if ppadbs.auto_detection_shadow_build_systems:
            names = ", ".join(
                sorted(x for x in ppadbs.auto_detection_shadow_build_systems)
            )
            _debug_log(f"Build system {ppadbs.manifest_keyword} excludes: {names}!")
        excludes.update(ppadbs.auto_detection_shadow_build_systems)
        options.append(ppadbs)

    if not options:
        _debug_log("Zero candidates; continuing without a build system")
        return None

    if excludes:
        names = ", ".join(sorted(x for x in excludes))
        _debug_log(f"The following build systems have been excluded: {names}!")
        remaining_options = [o for o in options if o.manifest_keyword not in excludes]
    else:
        remaining_options = options

    if len(remaining_options) > 1:
        names = ", ".join(o.manifest_keyword for o in remaining_options)
        # TODO: This means adding an auto-detectable build system to an existing plugin causes FTBFS
        # We need a better way of handling this. Probably the build systems should include
        # a grace timer based on d/changelog. Anything before the changelog date is in
        # "grace mode" and will not conflict with a build system that is. If all choices
        # are in "grace mode", "oldest one" wins.
        _error(
            f"Multiple build systems match, please pick one explicitly (under `builds:`): {names}"
        )

    if not remaining_options:
        names = ", ".join(o.build_system_rule_type.__name__ for o in options)
        # TODO: Detect at registration time
        _error(
            f"Multiple build systems matched but they all shadowed each other: {names}."
            f" There is a bug in at least one of them!"
        )

    chosen_build_system = remaining_options[0]
    environment = default_build_environment_only(manifest)
    bs = run_in_context_of_plugin_wrap_errors(
        chosen_build_system.plugin_metadata.plugin_name,
        chosen_build_system.constructor,
        {
            "environment": environment,
        },
        AttributePath.builtin_path(),
        manifest,
    )
    bs.auto_generated_stem = ""
    return bs



from typing import NotRequired, List, Any, TypedDict

from debputy.manifest_parser.tagging_types import (
    DebputyParsedContent,
    TypeMapping,
)
from debputy.manifest_parser.base_types import OctalMode
from debputy.manifest_parser.declarative_parser import ParserGenerator
from debputy.plugin.api.impl_types import KnownPackagingFileInfo


class PPFReferenceDocumentation(TypedDict):
    description: NotRequired[str]
    format_documentation_uris: NotRequired[List[str]]


class PackagerProvidedFileJsonDescription(DebputyParsedContent):
    stem: str
    installed_path: str
    default_mode: NotRequired[OctalMode]
    default_priority: NotRequired[int]
    allow_name_segment: NotRequired[bool]
    allow_architecture_segment: NotRequired[bool]
    reference_documentation: NotRequired[PPFReferenceDocumentation]


class ManifestVariableJsonDescription(DebputyParsedContent):
    name: str
    value: str
    reference_documentation: NotRequired[str]


class PluginJsonMetadata(DebputyParsedContent):
    api_compat_version: int
    module: NotRequired[str]
    plugin_initializer: NotRequired[str]
    packager_provided_files: NotRequired[List[Any]]
    manifest_variables: NotRequired[List[Any]]
    known_packaging_files: NotRequired[List[Any]]


def _initialize_plugin_metadata_parser_generator() -> ParserGenerator:
    pc = ParserGenerator()
    pc.register_mapped_type(
        TypeMapping(
            OctalMode,
            str,
            lambda v, ap, _: OctalMode.parse_filesystem_mode(v, ap),
        )
    )
    return pc


PLUGIN_METADATA_PARSER_GENERATOR = _initialize_plugin_metadata_parser_generator()
PLUGIN_METADATA_PARSER = PLUGIN_METADATA_PARSER_GENERATOR.generate_parser(
    PluginJsonMetadata
)
PLUGIN_PPF_PARSER = PLUGIN_METADATA_PARSER_GENERATOR.generate_parser(
    PackagerProvidedFileJsonDescription
)
PLUGIN_MANIFEST_VARS_PARSER = PLUGIN_METADATA_PARSER_GENERATOR.generate_parser(
    ManifestVariableJsonDescription
)
PLUGIN_KNOWN_PACKAGING_FILES_PARSER = PLUGIN_METADATA_PARSER_GENERATOR.generate_parser(
    KnownPackagingFileInfo
)



from typing import (
    Optional,
    Any,
    Tuple,
    Union,
    Sequence,
    Literal,
    get_args,
    get_origin,
    Container,
)

from debputy.highlevel_manifest import MANIFEST_YAML
from debputy.linting.lint_util import LintState
from debputy.lsp.lsp_features import (
    lint_diagnostics,
    lsp_standard_handler,
    lsp_hover,
    lsp_completer,
    SecondaryLanguage,
    LanguageDispatchRule,
)
from debputy.lsp.lsp_generic_yaml import (
    error_range_at_position,
    YAML_COMPLETION_HINT_KEY,
    insert_complete_marker_snippet,
    yaml_key_range,
    yaml_flag_unknown_key,
    _trace_cursor,
    generic_yaml_hover,
    resolve_keyword,
    DEBPUTY_PLUGIN_METADATA,
    maybe_quote_yaml_value,
    completion_from_attr,
)
from debputy.lsp.quickfixes import propose_correct_text_quick_fix
from debputy.lsp.vendoring._deb822_repro.locatable import (
    Position as TEPosition,
    Range as TERange,
)
from debputy.lsprotocol.types import (
    TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL,
    HoverParams,
    Hover,
    TEXT_DOCUMENT_CODE_ACTION,
    CompletionParams,
    CompletionList,
    CompletionItem,
    DiagnosticRelatedInformation,
    Location,
)
from debputy.manifest_parser.declarative_parser import (
    AttributeDescription,
    ParserGenerator,
    DeclarativeNonMappingInputParser,
)
from debputy.manifest_parser.declarative_parser import DeclarativeMappingInputParser
from debputy.manifest_parser.tagging_types import DebputyDispatchableType
from debputy.manifest_parser.util import AttributePath
from debputy.plugin.api.impl_types import (
    DeclarativeInputParser,
    DispatchingParserBase,
    ListWrappedDeclarativeInputParser,
    InPackageContextParser,
    DeclarativeValuelessKeywordInputParser,
    PluginProvidedParser,
)
from debputy.plugin.api.parser_tables import OPARSER_MANIFEST_ROOT
from debputy.plugin.api.spec import DebputyIntegrationMode
from debputy.plugin.debputy.private_api import Capability, load_libcap
from debputy.util import _info
from debputy.yaml.compat import (
    CommentedMap,
    CommentedSeq,
    MarkedYAMLError,
    YAMLError,
)

try:
    from pygls.server import LanguageServer
    from debputy.lsp.debputy_ls import DebputyLanguageServer
except ImportError:
    pass


_DISPATCH_RULE = LanguageDispatchRule.new_rule(
    "debian/debputy.manifest",
    "debian/debputy.manifest",
    [
        SecondaryLanguage("debputy.manifest"),
        # LSP's official language ID for YAML files
        SecondaryLanguage("yaml", filename_based_lookup=True),
    ],
)


lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_CODE_ACTION)
lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)


@lint_diagnostics(_DISPATCH_RULE)
async def _lint_debian_debputy_manifest(lint_state: LintState) -> None:
    lines = lint_state.lines
    try:
        content = MANIFEST_YAML.load("".join(lines))
    except MarkedYAMLError as e:
        if e.context_mark:
            line = e.context_mark.line
            column = e.context_mark.column
        else:
            line = e.problem_mark.line
            column = e.problem_mark.column
        error_range = error_range_at_position(
            lines,
            line,
            column,
        )
        lint_state.emit_diagnostic(
            error_range,
            f"YAML parse error: {e}",
            "error",
            "debputy",
        )
    except YAMLError as e:
        error_range = TERange(
            TEPosition(0, 0),
            TEPosition(0, len(lines[0])),
        )
        lint_state.emit_diagnostic(
            error_range,
            f"Unknown YAML parse error: {e} [{e!r}]",
            "error",
            "debputy",
        )
    else:
        feature_set = lint_state.plugin_feature_set
        pg = feature_set.manifest_parser_generator
        root_parser = pg.dispatchable_object_parsers[OPARSER_MANIFEST_ROOT]
        debputy_integration_mode = lint_state.debputy_metadata.debputy_integration_mode

        _lint_content(
            lint_state,
            pg,
            root_parser,
            debputy_integration_mode,
            content,
        )


def _integration_mode_allows_key(
    lint_state: LintState,
    debputy_integration_mode: Optional[DebputyIntegrationMode],
    expected_debputy_integration_modes: Optional[Container[DebputyIntegrationMode]],
    key: str,
    line: int,
    col: int,
) -> None:
    if debputy_integration_mode is None or expected_debputy_integration_modes is None:
        return
    if debputy_integration_mode in expected_debputy_integration_modes:
        return
    key_range = yaml_key_range(key, line, col)
    lint_state.emit_diagnostic(
        key_range,
        f'Feature "{key}" not supported in integration mode {debputy_integration_mode}',
        "error",
        "debputy",
    )


def _conflicting_key(
    lint_state: LintState,
    key_a: str,
    key_b: str,
    key_a_line: int,
    key_a_col: int,
    key_b_line: int,
    key_b_col: int,
) -> None:
    key_a_range = TERange(
        TEPosition(
            key_a_line,
            key_a_col,
        ),
        TEPosition(
            key_a_line,
            key_a_col + len(key_a),
        ),
    )
    key_b_range = TERange(
        TEPosition(
            key_b_line,
            key_b_col,
        ),
        TEPosition(
            key_b_line,
            key_b_col + len(key_b),
        ),
    )
    lint_state.emit_diagnostic(
        key_a_range,
        f'The "{key_a}" cannot be used with "{key_b}".',
        "error",
        "debputy",
        related_information=[
            lint_state.related_diagnostic_information(
                key_b_range, f'The attribute "{key_b}" is used here.'
            ),
        ],
    )

    lint_state.emit_diagnostic(
        key_b_range,
        f'The "{key_b}" cannot be used with "{key_a}".',
        "error",
        "debputy",
        related_information=[
            lint_state.related_diagnostic_information(
                key_a_range,
                f'The attribute "{key_a}" is used here.',
            ),
        ],
    )


def _remaining_line(lint_state: LintState, line_no: int, pos_start: int) -> "TERange":
    raw_line = lint_state.lines[line_no].rstrip()
    pos_end = len(raw_line)
    return TERange(
        TEPosition(
            line_no,
            pos_start,
        ),
        TEPosition(
            line_no,
            pos_end,
        ),
    )


def _lint_attr_value(
    lint_state: LintState,
    attr: AttributeDescription,
    pg: ParserGenerator,
    debputy_integration_mode: Optional[DebputyIntegrationMode],
    key: str,
    value: Any,
    pos: Tuple[int, int],
) -> None:
    target_attr_type = attr.attribute_type
    type_mapping = pg.get_mapped_type_from_target_type(target_attr_type)
    source_attr_type = target_attr_type
    if type_mapping is not None:
        source_attr_type = type_mapping.source_type
    orig = get_origin(source_attr_type)
    valid_values: Optional[Sequence[Any]] = None
    if orig == Literal:
        valid_values = get_args(attr.attribute_type)
    elif orig == bool or attr.attribute_type == bool:
        valid_values = (True, False)
    elif isinstance(target_attr_type, type):
        if issubclass(target_attr_type, Capability):
            has_libcap, _, is_valid_cap = load_libcap()
            if has_libcap and not is_valid_cap(value):
                line_no, cursor_pos = pos
                cap_range = _remaining_line(lint_state, line_no, cursor_pos)
                lint_state.emit_diagnostic(
                    cap_range,
                    "The value could not be parsed as a capability via cap_from_text on this system",
                    "warning",
                    "debputy",
                )
            return
        if issubclass(target_attr_type, DebputyDispatchableType):
            parser = pg.dispatch_parser_table_for(target_attr_type)
            _lint_content(
                lint_state,
                pg,
                parser,
                debputy_integration_mode,
                value,
            )
            return

    if valid_values is None or value in valid_values:
        return
    line_no, cursor_pos = pos
    value_range = _remaining_line(lint_state, line_no, cursor_pos)
    lint_state.emit_diagnostic(
        value_range,
        f'Not a supported value for "{key}"',
        "error",
        "debputy",
        quickfixes=[
            propose_correct_text_quick_fix(_as_yaml_value(m)) for m in valid_values
        ],
    )


def _as_yaml_value(v: Any) -> str:
    if isinstance(v, bool):
        return str(v).lower()
    return str(v)


def _lint_declarative_mapping_input_parser(
    lint_state: LintState,
    pg: ParserGenerator,
    parser: DeclarativeMappingInputParser,
    debputy_integration_mode: Optional[DebputyIntegrationMode],
    content: Any,
) -> None:
    if not isinstance(content, CommentedMap):
        return
    lc = content.lc
    for key, value in content.items():
        attr = parser.manifest_attributes.get(key)
        line, col = lc.key(key)
        if attr is None:
            corrected_key = yaml_flag_unknown_key(
                lint_state,
                key,
                parser.manifest_attributes,
                line,
                col,
            )
            if corrected_key:
                key = corrected_key
                attr = parser.manifest_attributes.get(corrected_key)
        if attr is None:
            continue

        _lint_attr_value(
            lint_state,
            attr,
            pg,
            debputy_integration_mode,
            key,
            value,
            lc.value(key),
        )

        for forbidden_key in attr.conflicting_attributes:
            if forbidden_key in content:
                con_line, con_col = lc.key(forbidden_key)
                _conflicting_key(
                    lint_state,
                    key,
                    forbidden_key,
                    line,
                    col,
                    con_line,
                    con_col,
                )
    for mx in parser.mutually_exclusive_attributes:
        matches = content.keys() & mx
        if len(matches) < 2:
            continue
        key, *others = list(matches)
        line, col = lc.key(key)
        for other in others:
            con_line, con_col = lc.key(other)
            _conflicting_key(
                lint_state,
                key,
                other,
                line,
                col,
                con_line,
                con_col,
            )


def _lint_content(
    lint_state: LintState,
    pg: ParserGenerator,
    parser: DeclarativeInputParser[Any],
    debputy_integration_mode: Optional[DebputyIntegrationMode],
    content: Any,
) -> None:
    if isinstance(parser, DispatchingParserBase):
        if not isinstance(content, CommentedMap):
            return
        lc = content.lc
        for key, value in content.items():
            is_known = parser.is_known_keyword(key)
            line, col = lc.key(key)
            orig_key = key
            if not is_known:
                corrected_key = yaml_flag_unknown_key(
                    lint_state,
                    key,
                    parser.registered_keywords(),
                    line,
                    col,
                    unknown_keys_diagnostic_severity=parser.unknown_keys_diagnostic_severity,
                )
                if corrected_key is not None:
                    key = corrected_key
                    is_known = True

            if is_known:
                subparser = parser.parser_for(key)
                assert subparser is not None
                _integration_mode_allows_key(
                    lint_state,
                    debputy_integration_mode,
                    subparser.parser.expected_debputy_integration_mode,
                    orig_key,
                    line,
                    col,
                )
                _lint_content(
                    lint_state,
                    pg,
                    subparser.parser,
                    debputy_integration_mode,
                    value,
                )
    elif isinstance(parser, ListWrappedDeclarativeInputParser):
        if not isinstance(content, CommentedSeq):
            return
        subparser = parser.delegate
        for value in content:
            _lint_content(lint_state, pg, subparser, debputy_integration_mode, value)
    elif isinstance(parser, InPackageContextParser):
        if not isinstance(content, CommentedMap):
            return
        known_packages = lint_state.binary_packages
        lc = content.lc
        for k, v in content.items():
            if k is None or (
                "{{" not in k and known_packages is not None and k not in known_packages
            ):
                line, col = lc.key(k)
                yaml_flag_unknown_key(
                    lint_state,
                    k,
                    known_packages,
                    line,
                    col,
                    message_format='Unknown package "{key}".',
                )
            _lint_content(
                lint_state,
                pg,
                parser.delegate,
                debputy_integration_mode,
                v,
            )
    elif isinstance(parser, DeclarativeMappingInputParser):
        _lint_declarative_mapping_input_parser(
            lint_state,
            pg,
            parser,
            debputy_integration_mode,
            content,
        )


def keywords_with_parser(
    parser: Union[DeclarativeMappingInputParser, DispatchingParserBase],
) -> Tuple[str, PluginProvidedParser]:
    for keyword in parser.registered_keywords():
        pp_subparser = parser.parser_for(keyword)
        yield keyword, pp_subparser


def completion_item(
    quoted_keyword: str,
    pp_subparser: PluginProvidedParser,
) -> CompletionItem:
    inline_reference_documentation = pp_subparser.parser.inline_reference_documentation
    synopsis = (
        inline_reference_documentation.synopsis
        if inline_reference_documentation
        else None
    )
    return CompletionItem(
        quoted_keyword,
        detail=synopsis,
    )


@lsp_completer(_DISPATCH_RULE)
def debputy_manifest_completer(
    ls: "DebputyLanguageServer",
    params: CompletionParams,
) -> Optional[Union[CompletionList, Sequence[CompletionItem]]]:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lines = doc.lines
    server_position = doc.position_codec.position_from_client_units(
        lines, params.position
    )
    orig_line = lines[server_position.line].rstrip()
    has_colon = ":" in orig_line
    added_key = insert_complete_marker_snippet(lines, server_position)
    attempts = 1 if added_key else 2
    content = None

    while attempts > 0:
        attempts -= 1
        try:
            content = MANIFEST_YAML.load("".join(lines))
            break
        except MarkedYAMLError as e:
            context_line = (
                e.context_mark.line if e.context_mark else e.problem_mark.line
            )
            if (
                e.problem_mark.line != server_position.line
                and context_line != server_position.line
            ):
                l_data = (
                    lines[e.problem_mark.line].rstrip()
                    if e.problem_mark.line < len(lines)
                    else "N/A (OOB)"
                )

                _info(f"Parse error on line: {e.problem_mark.line}: {l_data}")
                return None

            if attempts > 0:
                # Try to make it a key and see if that fixes the problem
                new_line = (
                    lines[server_position.line].rstrip() + YAML_COMPLETION_HINT_KEY
                )
                lines[server_position.line] = new_line
        except YAMLError:
            break
    if content is None:
        context = lines[server_position.line].replace("\n", "\\n")
        _info(f"Completion failed: parse error: Line in question: {context}")
        return None
    attribute_root_path = AttributePath.root_path(content)
    m = _trace_cursor(content, attribute_root_path, server_position)

    if m is None:
        _info("No match")
        return None
    matched_key, attr_path, matched, parent = m
    _info(f"Matched path: {matched} (path: {attr_path.path}) [{matched_key=}]")
    feature_set = ls.plugin_feature_set
    root_parser = feature_set.manifest_parser_generator.dispatchable_object_parsers[
        OPARSER_MANIFEST_ROOT
    ]
    segments = list(attr_path.path_segments())
    km = resolve_keyword(
        root_parser,
        DEBPUTY_PLUGIN_METADATA,
        segments,
        0,
        feature_set.manifest_parser_generator,
        is_completion_attempt=True,
    )
    if km is None:
        return None
    parser, _, at_depth_idx = km
    _info(f"Match leaf parser {at_depth_idx} -- {parser.__class__}")
    items = []
    if at_depth_idx + 1 >= len(segments):
        if isinstance(parser, DispatchingParserBase):
            if matched_key:
                items = [
                    completion_item(
                        (
                            maybe_quote_yaml_value(k)
                            if has_colon
                            else f"{maybe_quote_yaml_value(k)}:"
                        ),
                        pp_subparser,
                    )
                    for k, pp_subparser in keywords_with_parser(parser)
                    if k not in parent
                    and not isinstance(
                        pp_subparser.parser,
                        DeclarativeValuelessKeywordInputParser,
                    )
                ]
            else:
                items = [
                    completion_item(maybe_quote_yaml_value(k), pp_subparser)
                    for k, pp_subparser in keywords_with_parser(parser)
                    if k not in parent
                    and isinstance(
                        pp_subparser.parser,
                        DeclarativeValuelessKeywordInputParser,
                    )
                ]
        elif isinstance(parser, InPackageContextParser):
            binary_packages = ls.lint_state(doc).binary_packages
            if binary_packages is not None:
                items = [
                    CompletionItem(
                        maybe_quote_yaml_value(p)
                        if has_colon
                        else f"{maybe_quote_yaml_value(p)}:"
                    )
                    for p in binary_packages
                    if p not in parent
                ]
        elif isinstance(parser, DeclarativeMappingInputParser):
            if matched_key:
                _info("Match attributes")
                locked = set(parent)
                for mx in parser.mutually_exclusive_attributes:
                    if not mx.isdisjoint(parent.keys()):
                        locked.update(mx)
                for attr_name, attr in parser.manifest_attributes.items():
                    if not attr.conflicting_attributes.isdisjoint(parent.keys()):
                        locked.add(attr_name)
                        break
                items = [
                    CompletionItem(
                        maybe_quote_yaml_value(k)
                        if has_colon
                        else f"{maybe_quote_yaml_value(k)}:"
                    )
                    for k in parser.manifest_attributes
                    if k not in locked
                ]
            else:
                # Value
                key = segments[at_depth_idx] if len(segments) > at_depth_idx else None
                attr = parser.manifest_attributes.get(key)
                if attr is not None:
                    _info(f"Expand value / key: {key} -- {attr.attribute_type}")
                    items = completion_from_attr(
                        attr,
                        feature_set.manifest_parser_generator,
                        matched,
                    )
                else:
                    _info(
                        f"Expand value / key: {key} -- !! {list(parser.manifest_attributes)}"
                    )
        elif isinstance(parser, DeclarativeNonMappingInputParser):
            attr = parser.alt_form_parser
            items = completion_from_attr(
                attr,
                feature_set.manifest_parser_generator,
                matched,
            )
    return items


@lsp_hover(_DISPATCH_RULE)
def debputy_manifest_hover(
    ls: "DebputyLanguageServer",
    params: HoverParams,
) -> Optional[Hover]:
    return generic_yaml_hover(
        ls,
        params,
        lambda pg: pg.dispatchable_object_parsers[OPARSER_MANIFEST_ROOT],
    )



import functools
import os
from typing import Any, Tuple

from debputy.plugin.api import (
    DebputyPluginInitializer,
    BinaryCtrlAccessor,
    PackageProcessingContext,
)
from debputy.util import _error


def initialize(api: DebputyPluginInitializer) -> None:
    api.metadata_or_maintscript_detector(
        "numpy-depends",
        numpy3_versions,
        # Probably not necessary, but this is the most faithful conversion
        package_type=["deb", "udeb"],
    )


@functools.lru_cache
def _parse_numpy3_versions(versions_file: str) -> Tuple[str, str]:
    attributes = {}
    try:
        with open(versions_file, "rt", encoding="utf-8") as fd:
            for line in fd:
                if line.startswith("#") or line.isspace():
                    continue
                k, v = line.split()
                attributes[k] = v
    except FileNotFoundError:
        _error(
            f"Missing Build-Dependency on python3-numpy to ensure {versions_file}"
            " is present."
        )

    try:
        api_min_version = attributes["api-min-version"]
        abi_version = attributes["abi"]
    except KeyError as e:
        k = e.args[0]
        _error(f'Expected {versions_file} to contain the key "{k}"')
        assert False

    return api_min_version, abi_version


def numpy3_versions(
    _unused: Any,
    ctrl: BinaryCtrlAccessor,
    context: PackageProcessingContext,
) -> None:
    if context.binary_package.is_arch_all:
        dep = "python3-numpy"
    else:
        # Note we do not support --strict; codesearch.d.n suggests it is not used
        # anywhere and this saves us figuring out how to support it here.
        versions_file = os.environ.get("_NUMPY_TEST_PATH", "/usr/share/numpy3/versions")
        api_min_version, abi_version = _parse_numpy3_versions(versions_file)
        dep = f"python3-numpy (>= {api_min_version}), python3-numpy-abi{abi_version}"
    ctrl.substvars.add_dependency("python3:Depends", dep)



from .library import Library as Library

class Hex:
    def find(self, color: str | int) -> str: ...
    @staticmethod
    def cube(x: int) -> int: ...
    def fit(self, hex_val: str, ref: int) -> int: ...



import dataclasses
import functools
import importlib.resources
import re
import textwrap
from typing import (
    Type,
    TypeVar,
    Generic,
    Optional,
    List,
    Union,
    Callable,
    Mapping,
    Self,
    Dict,
    Iterable,
    Any,
    Tuple,
)

import debputy.lsp.data as data_dir
from debputy.lsp.lsp_reference_keyword import ALL_PUBLIC_NAMED_STYLES
from debputy.lsp.vendoring._deb822_repro.types import FormatterCallback
from debputy.lsp.vendoring.wrap_and_sort import wrap_and_sort_formatter
from debputy.packages import SourcePackage
from debputy.util import _error
from debputy.yaml import MANIFEST_YAML
from debputy.yaml.compat import CommentedMap

PT = TypeVar("PT", bool, str, int)


BUILTIN_STYLES = "maint-preferences.yaml"

_NORMALISE_FIELD_CONTENT_KEY = ["deb822", "normalize-field-content"]
_UPLOADER_SPLIT_RE = re.compile(r"(?<=>)\s*,")

_WAS_OPTIONS = {
    "-a": ("deb822_always_wrap", True),
    "--always-wrap": ("deb822_always_wrap", True),
    "-s": ("deb822_short_indent", True),
    "--short-indent": ("deb822_short_indent", True),
    "-t": ("deb822_trailing_separator", True),
    "--trailing-separator": ("deb822_trailing_separator", True),
    # Noise option for us; we do not accept `--no-keep-first` though
    "-k": (None, True),
    "--keep-first": (None, True),
    "--no-keep-first": ("DISABLE_NORMALIZE_STANZA_ORDER", True),
    "-b": ("deb822_normalize_stanza_order", True),
    "--sort-binary-packages": ("deb822_normalize_stanza_order", True),
}

_WAS_DEFAULTS = {
    "deb822_always_wrap": False,
    "deb822_short_indent": False,
    "deb822_trailing_separator": False,
    "deb822_normalize_stanza_order": False,
    "deb822_normalize_field_content": True,
}


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class PreferenceOption(Generic[PT]):
    key: Union[str, List[str]]
    expected_type: Union[Type[PT], Callable[[Any], Optional[str]]]
    description: str
    default_value: Optional[Union[PT, Callable[[CommentedMap], Optional[PT]]]] = None

    @property
    def name(self) -> str:
        if isinstance(self.key, str):
            return self.key
        return ".".join(self.key)

    @property
    def attribute_name(self) -> str:
        return self.name.replace("-", "_").replace(".", "_")

    def extract_value(
        self,
        filename: str,
        key: str,
        data: CommentedMap,
    ) -> Optional[PT]:
        v = data.mlget(self.key, list_ok=True)
        if v is None:
            default_value = self.default_value
            if callable(default_value):
                return default_value(data)
            return default_value
        val_issue: Optional[str] = None
        expected_type = self.expected_type
        if not isinstance(expected_type, type) and callable(self.expected_type):
            val_issue = self.expected_type(v)
        elif not isinstance(v, self.expected_type):
            val_issue = f"It should have been a {self.expected_type} but it was not"

        if val_issue is None:
            return v
        raise ValueError(
            f'The value "{self.name}" for key {key} in file "{filename}" was incorrect: {val_issue}'
        )


def _is_packaging_team_default(m: CommentedMap) -> bool:
    v = m.get("canonical-name")
    if not isinstance(v, str):
        return False
    v = v.lower()
    return v.endswith((" maintainer", " maintainers", " team"))


def _false_when_formatting_content(m: CommentedMap) -> Optional[bool]:
    return m.mlget(_NORMALISE_FIELD_CONTENT_KEY, list_ok=True, default=False) is True


MAINT_OPTIONS: List[PreferenceOption] = [
    PreferenceOption(
        key="canonical-name",
        expected_type=str,
        description=textwrap.dedent(
            """\
        Canonical spelling/case of the maintainer name.

        The `debputy` linter will emit a diagnostic if the name is not spelled exactly as provided here.
        Can be useful to ensure your name is updated after a change of name.
        """
        ),
    ),
    PreferenceOption(
        key="is-packaging-team",
        expected_type=bool,
        default_value=_is_packaging_team_default,
        description=textwrap.dedent(
            """\
    Whether this entry is for a packaging team

    This affects how styles are applied when multiple maintainers (`Maintainer` + `Uploaders`) are listed
    in `debian/control`. For package teams, the team preference prevails when the team is in the `Maintainer`
    field. For non-packaging teams, generally the rules do not apply as soon as there are co-maintainers.

    The default is derived from the canonical name. If said name ends with phrases like "Team" or "Maintainer"
    then the email is assumed to be for a team by default.
    """
        ),
    ),
    PreferenceOption(
        key="formatting",
        expected_type=lambda x: (
            None
            if isinstance(x, EffectiveFormattingPreference)
            else "It should have been a EffectiveFormattingPreference but it was not"
        ),
        default_value=None,
        description=textwrap.dedent(
            """\
    The formatting preference of the maintainer. Can either be a string for a named style or an inline
    style.
    """
        ),
    ),
]

FORMATTING_OPTIONS = [
    PreferenceOption(
        key=["deb822", "short-indent"],
        expected_type=bool,
        description=textwrap.dedent(
            """\
    Whether to use "short" indents for relationship fields (such as `Depends`).

    This roughly corresponds to `wrap-and-sort`'s `-s` option.

    **Example**:

    When `true`, the following:
    ```
    Depends: foo,
             bar
    ```

    would be reformatted as:

    ```
    Depends:
     foo,
     bar
    ```

    (Assuming `formatting.deb822.short-indent` is `false`)

    Note that defaults to `false` *if* (and only if) other formatting options will trigger reformat of
    the field and this option has not been set. Setting this option can trigger reformatting of fields
    that span multiple lines.

    Additionally, this only triggers when a field is being reformatted. Generally that requires
    another option such as `formatting.deb822.normalize-field-content` for that to happen.
    """
        ),
    ),
    PreferenceOption(
        key=["deb822", "always-wrap"],
        expected_type=bool,
        description=textwrap.dedent(
            """\
    Whether to always wrap fields (such as `Depends`).

    This roughly corresponds to `wrap-and-sort`'s `-a` option.

    **Example**:

    When `true`, the following:
    ```
    Depends: foo, bar
    ```

    would be reformatted as:

    ```
    Depends: foo,
             bar
    ```

    (Assuming `formatting.deb822.short-indent` is `false`)

    This option only applies to fields where formatting is a pure style preference. As an
    example, `Description` (`debian/control`) or `License` (`debian/copyright`) will not
    be affected by this option.

    Note: When `true`, this option overrules `formatting.deb822.max-line-length` when they interact. 
    Additionally, this only triggers when a field is being reformatted. Generally that requires
    another option such as `formatting.deb822.normalize-field-content` for that to happen.
    """
        ),
    ),
    PreferenceOption(
        key=["deb822", "trailing-separator"],
        expected_type=bool,
        default_value=False,
        description=textwrap.dedent(
            """\
    Whether to always end relationship fields (such as `Depends`) with a trailing separator.

    This roughly corresponds to `wrap-and-sort`'s `-t` option.

    **Example**:

    When `true`, the following:
    ```
    Depends: foo,
             bar
    ```

    would be reformatted as:

    ```
    Depends: foo,
             bar,
    ```

    Note: The trailing separator is only applied if the field is reformatted. This means this option
    generally requires another option to trigger reformatting (like
    `formatting.deb822.normalize-field-content`).
    """
        ),
    ),
    PreferenceOption(
        key=["deb822", "max-line-length"],
        expected_type=int,
        default_value=79,
        description=textwrap.dedent(
            """\
    How long a value line can be before it should be line wrapped.

    This roughly corresponds to `wrap-and-sort`'s `--max-line-length` option.

    This option only applies to fields where formatting is a pure style preference. As an
    example, `Description` (`debian/control`) or `License` (`debian/copyright`) will not
    be affected by this option.

    Note: When `formatting.deb822.always-wrap` is `true`, then this option will be overruled.
    Additionally, this only triggers when a field is being reformatted. Generally that requires
    another option such as `formatting.deb822.normalize-field-content` for that to happen.
    """
        ),
    ),
    PreferenceOption(
        key=_NORMALISE_FIELD_CONTENT_KEY,
        expected_type=bool,
        default_value=False,
        description=textwrap.dedent(
            """\
    Whether to normalize field content.

    This roughly corresponds to the subset of `wrap-and-sort` that normalizes field content
    like sorting and normalizing relations or sorting the architecture field.

    **Example**:

    When `true`, the following:
    ```
    Depends: foo,
             bar|baz
    ```

    would be reformatted as:

    ```
    Depends: bar | baz,
             foo,
    ```

    This causes affected fields to always be rewritten and therefore be sure that other options
    such as `formatting.deb822.short-indent` or `formatting.deb822.always-wrap` is set according
    to taste.

    Note: The field may be rewritten without this being set to `true`. As an example, the `always-wrap`
    option can trigger a field rewrite. However, in that case, the values (including any internal whitespace)
    are left as-is while the whitespace normalization between the values is still applied.
    """
        ),
    ),
    PreferenceOption(
        key=["deb822", "normalize-field-order"],
        expected_type=bool,
        default_value=False,
        description=textwrap.dedent(
            """\
    Whether to normalize field order in a stanza.

    There is no `wrap-and-sort` feature matching this.

    **Example**:

    When `true`, the following:
    ```
    Depends: bar
    Package: foo
    ```

    would be reformatted as:

    ```
    Depends: foo
    Package: bar
    ```

    The field order is not by field name but by a logic order defined in `debputy` based on existing
    conventions. The `deb822` format does not dictate any field order inside stanzas in general, so
    reordering of fields is generally safe.

    If a field of the first stanza is known to be a format discriminator such as the `Format' in
    `debian/copyright`, then it will be put first. Generally that matches existing convention plus
    it maximizes the odds that existing tools will correctly identify the file format.
    """
        ),
    ),
    PreferenceOption(
        key=["deb822", "normalize-stanza-order"],
        expected_type=bool,
        default_value=False,
        description=textwrap.dedent(
            """\
    Whether to normalize stanza order in a file.

    This roughly corresponds to `wrap-and-sort`'s `-kb` feature except this may apply to other deb822
    files.

    **Example**:

    When `true`, the following:
    ```
    Source: zzbar

    Package: zzbar

    Package: zzbar-util

    Package: libzzbar-dev

    Package: libzzbar2
    ```

    would be reformatted as:

    ```
    Source: zzbar

    Package: zzbar

    Package: libzzbar2

    Package: libzzbar-dev

    Package: zzbar-util
    ```

    Reordering will only performed when:
      1) There is a convention for a normalized order
      2) The normalization can be performed without changing semantics

    Note: This option only guards style/preference related re-ordering. It does not influence
    warnings about the order being semantic incorrect (which will still be emitted regardless
    of this setting).
    """
        ),
    ),
]


@dataclasses.dataclass(slots=True, frozen=True)
class EffectiveFormattingPreference:
    deb822_short_indent: Optional[bool] = None
    deb822_always_wrap: Optional[bool] = None
    deb822_trailing_separator: bool = False
    deb822_normalize_field_content: bool = False
    deb822_normalize_field_order: bool = False
    deb822_normalize_stanza_order: bool = False
    deb822_max_line_length: int = 79

    @classmethod
    def from_file(
        cls,
        filename: str,
        key: str,
        styles: CommentedMap,
    ) -> Self:
        attr = {}

        for option in FORMATTING_OPTIONS:
            if not hasattr(cls, option.attribute_name):
                continue
            value = option.extract_value(filename, key, styles)
            attr[option.attribute_name] = value
        return cls(**attr)  # type: ignore

    @classmethod
    def aligned_preference(
        cls,
        a: Optional["EffectiveFormattingPreference"],
        b: Optional["EffectiveFormattingPreference"],
    ) -> Optional["EffectiveFormattingPreference"]:
        if a is None or b is None:
            return None

        for option in MAINT_OPTIONS:
            attr_name = option.attribute_name
            if not hasattr(EffectiveFormattingPreference, attr_name):
                continue
            a_value = getattr(a, attr_name)
            b_value = getattr(b, attr_name)
            if a_value != b_value:
                return None
        return a

    def deb822_formatter(self) -> FormatterCallback:
        line_length = self.deb822_max_line_length
        return wrap_and_sort_formatter(
            1 if self.deb822_short_indent else "FIELD_NAME_LENGTH",
            trailing_separator=self.deb822_trailing_separator,
            immediate_empty_line=self.deb822_short_indent or False,
            max_line_length_one_liner=(0 if self.deb822_always_wrap else line_length),
        )

    def replace(self, /, **changes: Any) -> Self:
        return dataclasses.replace(self, **changes)


@dataclasses.dataclass(slots=True, frozen=True)
class MaintainerPreference:
    canonical_name: Optional[str] = None
    is_packaging_team: bool = False
    formatting: Optional[EffectiveFormattingPreference] = None

    @classmethod
    def from_file(
        cls,
        filename: str,
        key: str,
        styles: CommentedMap,
    ) -> Self:
        attr = {}

        for option in MAINT_OPTIONS:
            if not hasattr(cls, option.attribute_name):
                continue
            value = option.extract_value(filename, key, styles)
            attr[option.attribute_name] = value
        return cls(**attr)  # type: ignore


class MaintainerPreferenceTable:

    def __init__(
        self,
        named_styles: Mapping[str, EffectiveFormattingPreference],
        maintainer_preferences: Mapping[str, MaintainerPreference],
    ) -> None:
        self._named_styles = named_styles
        self._maintainer_preferences = maintainer_preferences

    @classmethod
    def load_preferences(cls) -> Self:
        named_styles: Dict[str, EffectiveFormattingPreference] = {}
        maintainer_preferences: Dict[str, MaintainerPreference] = {}

        path = importlib.resources.files(data_dir).joinpath(BUILTIN_STYLES)

        with path.open("r", encoding="utf-8") as fd:
            parse_file(named_styles, maintainer_preferences, str(path), fd)

        missing_keys = set(named_styles.keys()).difference(
            ALL_PUBLIC_NAMED_STYLES.keys()
        )
        if missing_keys:
            missing_styles = ", ".join(sorted(missing_keys))
            _error(
                f"The following named styles are public API but not present in the config file: {missing_styles}"
            )

        # TODO: Support fetching styles online to pull them in faster than waiting for a stable release.
        return cls(named_styles, maintainer_preferences)

    @property
    def named_styles(self) -> Mapping[str, EffectiveFormattingPreference]:
        return self._named_styles

    @property
    def maintainer_preferences(self) -> Mapping[str, MaintainerPreference]:
        return self._maintainer_preferences


def parse_file(
    named_styles: Dict[str, EffectiveFormattingPreference],
    maintainer_preferences: Dict[str, MaintainerPreference],
    filename: str,
    fd,
) -> None:
    content = MANIFEST_YAML.load(fd)
    if not isinstance(content, CommentedMap):
        raise ValueError(
            f'The file "{filename}" should be a YAML file with a single mapping at the root'
        )
    try:
        maintainer_rules = content["maintainer-rules"]
        if not isinstance(maintainer_rules, CommentedMap):
            raise KeyError("maintainer-rules") from None
    except KeyError:
        raise ValueError(
            f'The file "{filename}" should have a "maintainer-rules" key which must be a mapping.'
        )
    named_styles_raw = content.get("formatting")
    if named_styles_raw is None or not isinstance(named_styles_raw, CommentedMap):
        named_styles_raw = {}

    for style_name, content in named_styles_raw.items():
        style = EffectiveFormattingPreference.from_file(
            filename,
            style_name,
            content,
        )
        named_styles[style_name] = style

    for maintainer_email, maintainer_pref in maintainer_rules.items():
        if not isinstance(maintainer_pref, CommentedMap):
            line_no = maintainer_rules.lc.key(maintainer_email).line
            raise ValueError(
                f'The value for maintainer "{maintainer_email}" should have been a mapping,'
                f' but it is not. The problem entry is at line {line_no} in "{filename}"'
            )
        formatting = maintainer_pref.get("formatting")
        if isinstance(formatting, str):
            try:
                style = named_styles[formatting]
            except KeyError:
                line_no = maintainer_rules.lc.key(maintainer_email).line
                raise ValueError(
                    f'The maintainer "{maintainer_email}" requested the named style "{formatting}",'
                    f' but said style was not defined {filename}. The problem entry is at line {line_no} in "{filename}"'
                ) from None
            maintainer_pref["formatting"] = style
        elif formatting is not None:
            maintainer_pref["formatting"] = EffectiveFormattingPreference.from_file(
                filename,
                "formatting",
                formatting,
            )
        mp = MaintainerPreference.from_file(
            filename,
            maintainer_email,
            maintainer_pref,
        )

        maintainer_preferences[maintainer_email] = mp


@functools.lru_cache(64)
def extract_maint_email(maint: str) -> str:
    if not maint.endswith(">"):
        return ""

    try:
        idx = maint.index("<")
    except ValueError:
        return ""
    return maint[idx + 1 : -1]


def _parse_salsa_ci_boolean(value: Union[str, int, bool]) -> bool:
    if isinstance(value, str):
        return value in ("yes", "1", "true")
    elif not isinstance(value, (int, bool)):
        raise TypeError("Unsupported value")
    else:
        return value is True or value == 1


def _read_salsa_ci_wrap_and_sort_enabled(salsa_ci: Optional[CommentedMap]) -> bool:
    sentinel = object()
    disable_wrap_and_sort_raw = salsa_ci.mlget(
        ["variables", "SALSA_CI_DISABLE_WRAP_AND_SORT"],
        list_ok=True,
        default=sentinel,
    )

    if disable_wrap_and_sort_raw is sentinel:
        enable_wrap_and_sort_raw = salsa_ci.mlget(
            ["variables", "SALSA_CI_ENABLE_WRAP_AND_SORT"],
            list_ok=True,
            default=None,
        )
        if enable_wrap_and_sort_raw is None or not isinstance(
            enable_wrap_and_sort_raw, (str, int, bool)
        ):
            return False

        return _parse_salsa_ci_boolean(enable_wrap_and_sort_raw)
    if not isinstance(disable_wrap_and_sort_raw, (str, int, bool)):
        return False

    disable_wrap_and_sort = _parse_salsa_ci_boolean(disable_wrap_and_sort_raw)
    return not disable_wrap_and_sort


def determine_effective_preference(
    maint_preference_table: MaintainerPreferenceTable,
    source_package: Optional[SourcePackage],
    salsa_ci: Optional[CommentedMap],
) -> Tuple[Optional[EffectiveFormattingPreference], Optional[str], Optional[str]]:
    style = source_package.fields.get("X-Style") if source_package is not None else None
    if style is not None:
        if style not in ALL_PUBLIC_NAMED_STYLES:
            return None, None, "X-Style contained an unknown/unsupported style"
        return maint_preference_table.named_styles.get(style), "debputy reformat", None

    if salsa_ci and _read_salsa_ci_wrap_and_sort_enabled(salsa_ci):
        wrap_and_sort_options = salsa_ci.mlget(
            ["variables", "SALSA_CI_WRAP_AND_SORT_ARGS"],
            list_ok=True,
            default=None,
        )
        if wrap_and_sort_options is None:
            wrap_and_sort_options = ""
        elif not isinstance(wrap_and_sort_options, str):
            return (
                None,
                None,
                "The salsa-ci had a non-string option for wrap-and-sort",
            )
        detected_style = parse_salsa_ci_wrap_and_sort_args(wrap_and_sort_options)
        tool_w_args = f"wrap-and-sort {wrap_and_sort_options}".strip()
        if detected_style is None:
            msg = "One or more of the wrap-and-sort options in the salsa-ci file was not supported"
        else:
            msg = None
        return detected_style, tool_w_args, msg
    if source_package is None:
        return None, None, None

    maint = source_package.fields.get("Maintainer")
    if maint is None:
        return None, None, None
    maint_email = extract_maint_email(maint)
    maint_pref = maint_preference_table.maintainer_preferences.get(maint_email)
    # Special-case "@packages.debian.org" when missing, since they are likely to be "ad-hoc"
    # teams that will not be registered. In that case, we fall back to looking at the uploader
    # preferences as-if the maintainer had not been listed at all.
    if maint_pref is None and not maint_email.endswith("@packages.debian.org"):
        return None, None, None
    if maint_pref is not None and maint_pref.is_packaging_team:
        # When the maintainer is registered as a packaging team, then we assume the packaging
        # team's style applies unconditionally.
        effective = maint_pref.formatting
        tool_w_args = _guess_tool_from_style(maint_preference_table, effective)
        return effective, tool_w_args, None
    uploaders = source_package.fields.get("Uploaders")
    if uploaders is None:
        detected_style = maint_pref.formatting if maint_pref is not None else None
        tool_w_args = _guess_tool_from_style(maint_preference_table, detected_style)
        return detected_style, tool_w_args, None
    all_styles: List[Optional[EffectiveFormattingPreference]] = []
    if maint_pref is not None:
        all_styles.append(maint_pref.formatting)
    for uploader in _UPLOADER_SPLIT_RE.split(uploaders):
        uploader_email = extract_maint_email(uploader)
        uploader_pref = maint_preference_table.maintainer_preferences.get(
            uploader_email
        )
        all_styles.append(uploader_pref.formatting if uploader_pref else None)

    if not all_styles:
        return None, None, None
    r = functools.reduce(EffectiveFormattingPreference.aligned_preference, all_styles)
    assert not isinstance(r, MaintainerPreference)
    tool_w_args = _guess_tool_from_style(maint_preference_table, r)
    return r, tool_w_args, None


def _guess_tool_from_style(
    maint_preference_table: MaintainerPreferenceTable,
    pref: Optional[EffectiveFormattingPreference],
) -> Optional[str]:
    if pref is None:
        return None
    if maint_preference_table.named_styles["black"] == pref:
        return "debputy reformat"
    return None


def _split_options(args: Iterable[str]) -> Iterable[str]:
    for arg in args:
        if arg.startswith("--"):
            yield arg
            continue
        if not arg.startswith("-") or len(arg) < 2:
            yield arg
            continue
        for sarg in arg[1:]:
            yield f"-{sarg}"


@functools.lru_cache
def parse_salsa_ci_wrap_and_sort_args(
    args: str,
) -> Optional[EffectiveFormattingPreference]:
    options = dict(_WAS_DEFAULTS)
    for arg in _split_options(args.split()):
        v = _WAS_OPTIONS.get(arg)
        if v is None:
            return None
        varname, value = v
        if varname is None:
            continue
        options[varname] = value
    if "DISABLE_NORMALIZE_STANZA_ORDER" in options:
        del options["DISABLE_NORMALIZE_STANZA_ORDER"]
        options["deb822_normalize_stanza_order"] = False

    return EffectiveFormattingPreference(**options)  # type: ignore



from debputy.plugin.api import (
    DebputyPluginInitializer,
    VirtualPath,
    BinaryCtrlAccessor,
    PackageProcessingContext,
)


def _udeb_metadata_detector(
    _path: VirtualPath,
    ctrl: BinaryCtrlAccessor,
    _context: PackageProcessingContext,
) -> None:
    ctrl.substvars["Test:Udeb-Metadata-Detector"] = "was-run"


def custom_plugin_initializer(api: DebputyPluginInitializer) -> None:
    api.packager_provided_file(
        "my-file",
        "/no-where/this/is/a/test/plugin/{name}.conf",
        post_formatting_rewrite=lambda x: x.replace("+", "_"),
    )
    api.metadata_or_maintscript_detector(
        "udeb-only", _udeb_metadata_detector, package_type="udeb"
    )
